from typing import Any

import autograd.numpy as anp
import pydantic.v1 as pd

from tidy3d.components.monitor import ModeMonitor
from tidy3d.components.types import Direction, FreqArray
from tidy3d.exceptions import ValidationError

from .types import NumberType
from .variables import Variable


class Metric(Variable):
    """
    Base class for all metrics.

    To subclass Metric, you must implement an evaluate() method that takes a SimulationData
    object and returns a scalar value.
    """

    def __repr__(self) -> str:
        return f'{self.type}("{self.monitor_name}")'


class ModeAmp(Metric):
    """
    Metric for calculating the mode coefficient from a ModeMonitor.

    Parameters
    ----------
    monitor_name : str
        The name of the mode monitor.
    freqs : FreqArray
        The frequency array.
    direction : Direction = "+"
        The direction of the mode.
    mode_index : pd.NonNegativeInt = 0
        The index of the mode.

    Examples
    --------
    >>> monitor = ModeMonitor(name="monitor1", freqs=[1.0])
    >>> mode_coeff = ModeAmp.from_mode_monitor(monitor)
    >>> data = SimulationData()  # Assume this is a valid SimulationData object
    >>> result = mode_coeff.evaluate(data)
    """

    monitor_name: str
    freqs: FreqArray
    direction: Direction = "+"
    mode_index: pd.NonNegativeInt = 0

    @pd.validator("freqs", always=True)
    def _single_frequency(cls, val: FreqArray) -> FreqArray:
        if len(val) != 1:
            raise ValidationError("Only a single frequency is supported at the moment.")
        return val

    @classmethod
    def from_mode_monitor(cls, monitor: ModeMonitor):
        return cls(monitor_name=monitor.name, freqs=monitor.freqs, mode_index=0)

    def evaluate(self, *args: Any, **kwargs: Any) -> NumberType:
        data = super().evaluate(*args, **kwargs)
        amps = (
            data[self.monitor_name]
            .amps.sel(direction=self.direction, mode_index=self.mode_index)
            .isel(f=0)
        )
        return anp.squeeze(amps.values.tolist())


class ModePower(ModeAmp):
    """
    Metric for calculating the mode power from a ModeMonitor.

    Examples
    --------
    >>> monitor = ModeMonitor(name="monitor1", freqs=[1.0])
    >>> mode_power = ModePower.from_mode_monitor(monitor)
    >>> data = SimulationData()  # Assume this is a valid SimulationData object
    >>> result = mode_power.evaluate(data)
    """

    def evaluate(self, *args: Any, **kwargs: Any) -> NumberType:
        amps = super().evaluate(*args, **kwargs)
        return abs(amps) ** 2



"""Tool for generating an S matrix automatically from a Tidy3d simulation and lumped port definitions."""

from __future__ import annotations

from typing import Dict, Tuple, Union

import numpy as np
import pydantic.v1 as pd

from ....components.base import cached_property
from ....components.data.sim_data import SimulationData
from ....components.geometry.utils_2d import snap_coordinate_to_grid
from ....components.simulation import Simulation
from ....components.source import GaussianPulse
from ....components.types import Ax
from ....components.viz import add_ax_if_none, equal_aspect
from ....constants import C_0, fp_eps
from ....exceptions import ValidationError
from ....web.api.container import BatchData
from ..ports.base_lumped import LumpedPortDataArray
from ..ports.coaxial_lumped import CoaxialLumpedPort
from ..ports.rectangular_lumped import LumpedPort
from .base import FWIDTH_FRAC, AbstractComponentModeler


class TerminalComponentModeler(AbstractComponentModeler):
    """Tool for modeling two-terminal multiport devices and computing port parameters
    with lumped ports."""

    ports: Tuple[Union[LumpedPort, CoaxialLumpedPort], ...] = pd.Field(
        (),
        title="Lumped ports",
        description="Collection of lumped ports associated with the network. "
        "For each port, one simulation will be run with a lumped port source.",
    )

    @equal_aspect
    @add_ax_if_none
    def plot_sim(
        self, x: float = None, y: float = None, z: float = None, ax: Ax = None, **kwargs
    ) -> Ax:
        """Plot a :class:`.Simulation` with all sources added for each port, for troubleshooting."""

        plot_sources = []
        for port_source in self.ports:
            source_0 = port_source.to_source(
                self._source_time, snap_center=None, grid=self.simulation.grid
            )
            plot_sources.append(source_0)
        sim_plot = self.simulation.copy(update=dict(sources=plot_sources))
        return sim_plot.plot(x=x, y=y, z=z, ax=ax, **kwargs)

    @equal_aspect
    @add_ax_if_none
    def plot_sim_eps(
        self, x: float = None, y: float = None, z: float = None, ax: Ax = None, **kwargs
    ) -> Ax:
        """Plot permittivity of the :class:`.Simulation` with all sources added for each port."""

        plot_sources = []
        for port_source in self.ports:
            source_0 = port_source.to_source(
                self._source_time, snap_center=None, grid=self.simulation.grid
            )
            plot_sources.append(source_0)
        sim_plot = self.simulation.copy(update=dict(sources=plot_sources))
        return sim_plot.plot_eps(x=x, y=y, z=z, ax=ax, **kwargs)

    @cached_property
    def sim_dict(self) -> Dict[str, Simulation]:
        """Generate all the :class:`.Simulation` objects for the port parameter calculation."""

        sim_dict = {}

        port_voltage_monitors = [
            port.to_voltage_monitor(self.freqs, snap_center=None) for port in self.ports
        ]
        port_current_monitors = [
            port.to_current_monitor(self.freqs, snap_center=None) for port in self.ports
        ]
        lumped_resistors = [port.to_load(snap_center=None) for port in self.ports]

        # Create a mesh override for each port in case refinement is needed.
        # The port is a flat surface, but when computing the port current,
        # we'll eventually integrate the magnetic field just above and below
        # this surface, so the mesh override needs to ensure that the mesh
        # is fine enough not only in plane, but also in the normal direction.
        # So in the normal direction, we'll make sure there are at least
        # 2 cell layers above and below whose size is the same as the in-plane
        # cell size in the override region. Also, to ensure that the port itself
        # is aligned with a grid boundary in the normal direction, two separate
        # override regions are defined, one above and one below the analytical
        # port region.
        mesh_overrides = []
        for port, lumped_resistor in zip(self.ports, lumped_resistors):
            if port.num_grid_cells:
                mesh_overrides.extend(lumped_resistor.to_mesh_overrides())

        new_mnts = list(self.simulation.monitors) + port_voltage_monitors + port_current_monitors

        # also, use the highest frequency in the simulation to define the grid, rather than the
        # source's central frequency, to ensure an accurate solution over the entire range
        grid_spec = self.simulation.grid_spec.copy(
            update={
                "wavelength": C_0 / np.max(self.freqs),
                "override_structures": list(self.simulation.grid_spec.override_structures)
                + mesh_overrides,
            }
        )

        # Checking if snapping is required needs the simulation to be created, because the
        # elements may impact the final grid discretization
        snap_and_recreate = False
        snap_centers = []
        for port in self.ports:
            update_dict = dict(
                monitors=new_mnts,
                lumped_elements=lumped_resistors,
                grid_spec=grid_spec,
            )

            sim_copy = self.simulation.copy(update=update_dict)
            port_source = port.to_source(self._source_time, snap_center=None, grid=sim_copy.grid)
            sim_copy = sim_copy.updated_copy(sources=[port_source])
            task_name = self._task_name(port=port)
            sim_dict[task_name] = sim_copy

            # Check if snapping to grid is needed
            port_center_on_axis = port.center[port.injection_axis]
            new_port_center = snap_coordinate_to_grid(
                sim_copy.grid, port_center_on_axis, port.injection_axis
            )
            snap_centers.append(new_port_center)
            if not np.isclose(port_center_on_axis, new_port_center, fp_eps, fp_eps):
                snap_and_recreate = True

        # Check if snapping was needed and if it was recreate the simulations
        if snap_and_recreate:
            sim_dict.clear()
            port_voltage_monitors = [
                port.to_voltage_monitor(self.freqs, snap_center=center)
                for port, center in zip(self.ports, snap_centers)
            ]
            port_current_monitors = [
                port.to_current_monitor(self.freqs, snap_center=center)
                for port, center in zip(self.ports, snap_centers)
            ]
            lumped_resistors = [
                port.to_load(snap_center=center) for port, center in zip(self.ports, snap_centers)
            ]
            new_mnts = (
                list(self.simulation.monitors) + port_voltage_monitors + port_current_monitors
            )
            for port, center in zip(self.ports, snap_centers):
                update_dict = dict(
                    monitors=new_mnts,
                    lumped_elements=lumped_resistors,
                    grid_spec=grid_spec,
                )

                sim_copy = self.simulation.copy(update=update_dict)
                port_source = port.to_source(
                    self._source_time, snap_center=center, grid=sim_copy.grid
                )
                sim_copy = sim_copy.updated_copy(sources=[port_source])
                task_name = self._task_name(port=port)
                sim_dict[task_name] = sim_copy

        # Check final simulations for grid size at ports
        for _, sim in sim_dict.items():
            TerminalComponentModeler._check_grid_size_at_ports(sim, self.ports)
        return sim_dict

    @cached_property
    def _source_time(self):
        """Helper to create a time domain pulse for the frequeny range of interest."""
        freq0 = np.mean(self.freqs)
        fdiff = max(self.freqs) - min(self.freqs)
        fwidth = max(fdiff, freq0 * FWIDTH_FRAC)
        return GaussianPulse(
            freq0=freq0, fwidth=fwidth, remove_dc_component=self.remove_dc_component
        )

    def _construct_smatrix(self, batch_data: BatchData) -> LumpedPortDataArray:
        """Post process ``BatchData`` to generate scattering matrix."""

        port_names = [port.name for port in self.ports]

        values = np.zeros(
            (len(port_names), len(port_names), len(self.freqs)),
            dtype=complex,
        )
        coords = dict(
            port_out=port_names,
            port_in=port_names,
            f=np.array(self.freqs),
        )
        a_matrix = LumpedPortDataArray(values, coords=coords)
        b_matrix = a_matrix.copy(deep=True)

        def port_ab(port: Union[LumpedPort, CoaxialLumpedPort], sim_data: SimulationData):
            """Helper to compute the port incident and reflected power waves."""
            voltage = port.compute_voltage(sim_data)
            current = port.compute_current(sim_data)

            power_a = (voltage + port.impedance * current) / 2 / np.sqrt(np.real(port.impedance))
            power_b = (voltage - port.impedance * current) / 2 / np.sqrt(np.real(port.impedance))
            return power_a, power_b

        # loop through source ports
        for port_in in self.ports:
            sim_data = batch_data[self._task_name(port=port_in)]

            for port_out in self.ports:
                a_out, b_out = port_ab(port_out, sim_data)
                a_matrix.loc[
                    dict(
                        port_in=port_in.name,
                        port_out=port_out.name,
                    )
                ] = a_out

                b_matrix.loc[
                    dict(
                        port_in=port_in.name,
                        port_out=port_out.name,
                    )
                ] = b_out

        s_matrix = self.ab_to_s(a_matrix, b_matrix)

        return s_matrix

    @pd.validator("simulation")
    def _validate_3d_simulation(cls, val):
        """Error if :class:`.Simulation` is not a 3D simulation"""

        if val.size.count(0.0) > 0:
            raise ValidationError(
                f"'{cls.__name__}' must be setup with a 3D simulation with all sizes greater than 0."
            )
        return val

    @staticmethod
    def _check_grid_size_at_ports(
        simulation: Simulation, ports: list[Union[LumpedPort, CoaxialLumpedPort]]
    ):
        """Raises :class:`.SetupError` if the grid is too coarse at port locations"""
        yee_grid = simulation.grid.yee
        for port in ports:
            port._check_grid_size(yee_grid)



# utilities for autograd derivative passing
from __future__ import annotations

import numpy as np
import pydantic.v1 as pd
import xarray as xr

from ..base import Tidy3dBaseModel
from ..data.data_array import ScalarFieldDataArray
from ..types import Bound, tidycomplex
from .types import PathType
from .utils import get_static

# we do this because importing these creates circular imports
FieldData = dict[str, ScalarFieldDataArray]
PermittivityData = dict[str, ScalarFieldDataArray]


class DerivativeInfo(Tidy3dBaseModel):
    """Stores derivative information passed to the ``.compute_derivatives`` methods."""

    paths: list[PathType] = pd.Field(
        ...,
        title="Paths to Traced Fields",
        description="List of paths to the traced fields that need derivatives calculated.",
    )

    E_der_map: FieldData = pd.Field(
        ...,
        title="Electric Field Gradient Map",
        description='Dataset where the field components ``("Ex", "Ey", "Ez")`` store the '
        "multiplication of the forward and adjoint electric fields. The tangential components "
        "of this dataset is used when computing adjoint gradients for shifting boundaries. "
        "All components are used when computing volume-based gradients.",
    )

    D_der_map: FieldData = pd.Field(
        ...,
        title="Displacement Field Gradient Map",
        description='Dataset where the field components ``("Ex", "Ey", "Ez")`` store the '
        "multiplication of the forward and adjoint displacement fields. The normal component "
        "of this dataset is used when computing adjoint gradients for shifting boundaries.",
    )

    E_fwd: FieldData = pd.Field(
        ...,
        title="Forward Electric Fields",
        description='Dataset where the field components ``("Ex", "Ey", "Ez")`` represent the '
        "forward electric fields used for computing gradients for a given structure.",
    )

    E_adj: FieldData = pd.Field(
        ...,
        title="Adjoint Electric Fields",
        description='Dataset where the field components ``("Ex", "Ey", "Ez")`` represent the '
        "adjoint electric fields used for computing gradients for a given structure.",
    )

    D_fwd: FieldData = pd.Field(
        ...,
        title="Forward Displacement Fields",
        description='Dataset where the field components ``("Ex", "Ey", "Ez")`` represent the '
        "forward displacement fields used for computing gradients for a given structure.",
    )

    D_adj: FieldData = pd.Field(
        ...,
        title="Adjoint Displacement Fields",
        description='Dataset where the field components ``("Ex", "Ey", "Ez")`` represent the '
        "adjoint displacement fields used for computing gradients for a given structure.",
    )

    eps_data: PermittivityData = pd.Field(
        ...,
        title="Permittivity Dataset",
        description="Dataset of relative permittivity values along all three dimensions. "
        "Used for automatically computing permittivity inside or outside of a simple geometry.",
    )

    eps_in: tidycomplex = pd.Field(
        title="Permittivity Inside",
        description="Permittivity inside of the ``Structure``. "
        "Typically computed from ``Structure.medium.eps_model``."
        "Used when it can not be computed from ``eps_data`` or when ``eps_approx==True``.",
    )

    eps_out: tidycomplex = pd.Field(
        ...,
        title="Permittivity Outside",
        description="Permittivity outside of the ``Structure``. "
        "Typically computed from ``Simulation.medium.eps_model``."
        "Used when it can not be computed from ``eps_data`` or when ``eps_approx==True``.",
    )

    bounds: Bound = pd.Field(
        ...,
        title="Geometry Bounds",
        description="Bounds corresponding to the structure, used in ``Medium`` calculations.",
    )

    frequency: float = pd.Field(
        ...,
        title="Frequency of adjoint simulation",
        description="Frequency at which the adjoint gradient is computed.",
    )

    eps_approx: bool = pd.Field(
        False,
        title="Use Permittivity Approximation",
        description="If ``True``, approximates outside permittivity using ``Simulation.medium``"
        "and the inside permittivity using ``Structure.medium``. "
        "Only set ``True`` for ``GeometryGroup`` handling where it is difficult to automatically "
        "evaluate the inside and outside relative permittivity for each geometry.",
    )

    def updated_paths(self, paths: list[PathType]) -> DerivativeInfo:
        """Update this ``DerivativeInfo`` with new set of paths."""
        return self.updated_copy(paths=paths)

    def grad_in_bases(self, spatial_coords: np.ndarray, basis_vectors: dict) -> dict:
        """Get the ``D_norm``, ``E_edge`` ``E_slab`` components of the gradient contributions."""

        # unpack electric and displacement fields
        E_fwd = self.E_fwd
        E_adj = self.E_adj
        D_fwd = self.D_fwd
        D_adj = self.D_adj

        # compute the E and D fields at the edge centers
        E_fwd_at_coords = self.evaluate_flds_at(fld_dataset=E_fwd, spatial_coords=spatial_coords)
        E_adj_at_coords = self.evaluate_flds_at(fld_dataset=E_adj, spatial_coords=spatial_coords)
        D_fwd_at_coords = self.evaluate_flds_at(fld_dataset=D_fwd, spatial_coords=spatial_coords)
        D_adj_at_coords = self.evaluate_flds_at(fld_dataset=D_adj, spatial_coords=spatial_coords)

        # project the relevant field quantities into their respective basis for gradient calculation
        D_fwd_norm = self.project_in_basis(D_fwd_at_coords, basis_vector=basis_vectors["norm"])
        D_adj_norm = self.project_in_basis(D_adj_at_coords, basis_vector=basis_vectors["norm"])

        E_fwd_perp1 = self.project_in_basis(E_fwd_at_coords, basis_vector=basis_vectors["perp1"])
        E_adj_perp1 = self.project_in_basis(E_adj_at_coords, basis_vector=basis_vectors["perp1"])

        E_fwd_perp2 = self.project_in_basis(E_fwd_at_coords, basis_vector=basis_vectors["perp2"])
        E_adj_perp2 = self.project_in_basis(E_adj_at_coords, basis_vector=basis_vectors["perp2"])

        # multiply forward and adjoint
        D_der_norm = D_fwd_norm * D_adj_norm
        E_der_perp1 = E_fwd_perp1 * E_adj_perp1
        E_der_perp2 = E_fwd_perp2 * E_adj_perp2

        return dict(D_norm=D_der_norm, E_perp1=E_der_perp1, E_perp2=E_der_perp2)

    @staticmethod
    def evaluate_flds_at(
        fld_dataset: dict[str, ScalarFieldDataArray],
        spatial_coords: np.ndarray,  # (N, 3)
    ) -> dict[str, ScalarFieldDataArray]:
        """Compute the value of an dict with keys Ex, Ey, Ez at a set of spatial locations."""

        xs, ys, zs = spatial_coords.T
        edge_index_dim = "edge_index"

        interp_kwargs = {}
        for dim, locations_dim in zip("xyz", (xs, ys, zs)):
            # only include dims where the data has more than 1 coord, to avoid warnings and errors
            if True or all(np.array(fld.coords).size > 1 for fld in fld_dataset.values()):
                interp_kwargs[dim] = xr.DataArray(locations_dim, dims=edge_index_dim)

        components = {}
        for fld_name, arr in fld_dataset.items():
            components[fld_name] = arr.interp(**interp_kwargs).sum("f")

        return components

    @staticmethod
    def project_in_basis(
        der_dataset: xr.Dataset,
        basis_vector: np.ndarray,
    ) -> xr.DataArray:
        """Project a derivative dataset along a supplied basis vector."""

        value = 0.0
        for coeffs, dim in zip(basis_vector.T, "xyz"):
            value += coeffs * der_dataset[f"E{dim}"]
        return value


# TODO: could we move this into a DataArray method?
def integrate_within_bounds(arr: xr.DataArray, dims: list[str], bounds: Bound) -> xr.DataArray:
    """integrate a data array within bounds, assumes bounds are [2, N] for N dims."""

    _arr = arr.copy()

    # order bounds with dimension first (N, 2)
    bounds = np.array(bounds).T

    all_coords = {}

    # loop over all dimensions
    for dim, (bmin, bmax) in zip(dims, bounds):
        bmin = get_static(bmin)
        bmax = get_static(bmax)

        coord_values = _arr.coords[dim].values

        # reset all coordinates outside of bounds to the bounds, so that dL = 0 in integral
        coord_values[coord_values < bmin] = bmin
        coord_values[coord_values > bmax] = bmax

        all_coords[dim] = coord_values

    _arr = _arr.assign_coords(**all_coords)

    # uses trapezoidal rule
    # https://docs.xarray.dev/en/stable/generated/xarray.DataArray.integrate.html
    dims_integrate = [dim for dim in dims if len(_arr.coords[dim]) > 1]
    return _arr.integrate(coord=dims_integrate)


__all__ = [
    "integrate_within_bounds",
    "DerivativeInfo",
]



"""Lumped port specialization with a rectangular geometry."""

import numpy as np
import pydantic.v1 as pd

from ....components.base import cached_property
from ....components.data.data_array import FreqDataArray
from ....components.data.sim_data import SimulationData
from ....components.geometry.base import Box
from ....components.geometry.utils_2d import increment_float
from ....components.grid.grid import Grid, YeeGrid
from ....components.lumped_element import LumpedResistor
from ....components.monitor import FieldMonitor
from ....components.source import GaussianPulse, UniformCurrentSource
from ....components.types import Axis, FreqArray
from ....components.validators import assert_plane
from ....constants import fp_eps
from ....exceptions import SetupError, ValidationError
from ...microwave import CurrentIntegralAxisAligned, VoltageIntegralAxisAligned
from .base_lumped import AbstractLumpedPort


class LumpedPort(AbstractLumpedPort, Box):
    """Class representing a single rectangular lumped port

    Example
    -------
    >>> port1 = LumpedPort(center=(0, 0, 0),
    ...             size=(0, 1, 2),
    ...             voltage_axis=2,
    ...             name="port_1",
    ...             impedance=50
    ...         )
    """

    voltage_axis: Axis = pd.Field(
        ...,
        title="Voltage Integration Axis",
        description="Specifies the axis along which the E-field line integral is performed when "
        "computing the port voltage. The integration axis must lie in the plane of the port.",
    )

    _plane_validator = assert_plane()

    @cached_property
    def injection_axis(self):
        """Injection axis of the port."""
        return self.size.index(0.0)

    @pd.validator("voltage_axis", always=True)
    def _voltage_axis_in_plane(cls, val, values):
        """Ensure voltage integration axis is in the port's plane."""
        size = values.get("size")
        if val == size.index(0.0):
            raise ValidationError("'voltage_axis' must lie in the port's plane.")
        return val

    @cached_property
    def current_axis(self) -> Axis:
        """Integration axis for computing the port current via the magnetic field."""
        return 3 - self.injection_axis - self.voltage_axis

    def to_source(
        self, source_time: GaussianPulse, snap_center: float, grid: Grid
    ) -> UniformCurrentSource:
        """Create a current source from the lumped port."""
        # Discretized source amps are manually zeroed out later if they
        # fall on Yee grid locations outside the analytical source region.
        component = "xyz"[self.voltage_axis]
        center = list(self.center)
        if snap_center:
            center[self.injection_axis] = snap_center
        return UniformCurrentSource(
            center=center,
            size=self.size,
            source_time=source_time,
            polarization=f"E{component}",
            name=self.name,
            interpolate=True,
            confine_to_bounds=True,
        )

    def to_load(self, snap_center: float) -> LumpedResistor:
        """Create a load resistor from the lumped port."""
        # 2D materials are currently snapped to the grid, so snapping here is not needed.
        # It is done here so plots of the simulation will more accurately portray the setup
        center = list(self.center)
        if snap_center:
            center[self.injection_axis] = snap_center
        return LumpedResistor(
            center=center,
            size=self.size,
            num_grid_cells=self.num_grid_cells,
            resistance=np.real(self.impedance),
            name=f"{self.name}_resistor",
            voltage_axis=self.voltage_axis,
        )

    def to_voltage_monitor(self, freqs: FreqArray, snap_center: float) -> FieldMonitor:
        """Field monitor to compute port voltage."""
        center = list(self.center)
        if snap_center:
            center[self.injection_axis] = snap_center

        e_component = "xyz"[self.voltage_axis]
        # Size of voltage monitor can essentially be 1D from ground to signal conductor
        voltage_mon_size = list(self.size)
        voltage_mon_size[self.injection_axis] = 0.0
        voltage_mon_size[self.current_axis] = 0.0
        # Create a voltage monitor
        return FieldMonitor(
            center=center,
            size=voltage_mon_size,
            freqs=freqs,
            fields=[f"E{e_component}"],
            name=self._voltage_monitor_name,
            colocate=False,
        )

    def to_current_monitor(self, freqs: FreqArray, snap_center: float) -> FieldMonitor:
        """Field monitor to compute port current."""
        center = list(self.center)
        if snap_center:
            center[self.injection_axis] = snap_center

        h_component = "xyz"[self.current_axis]
        h_cap_component = "xyz"[self.injection_axis]
        # Size of current monitor needs to encompass the current carrying 2D sheet
        # Needs to have a nonzero thickness so a closed loop of gridpoints around the 2D sheet can be formed
        dl = 2 * (increment_float(center[self.injection_axis], 1.0) - center[self.injection_axis])
        current_mon_size = list(self.size)
        current_mon_size[self.injection_axis] = dl
        current_mon_size[self.voltage_axis] = 0.0
        # Create a current monitor
        return FieldMonitor(
            center=center,
            size=current_mon_size,
            freqs=freqs,
            fields=[f"H{h_component}", f"H{h_cap_component}"],
            name=self._current_monitor_name,
            colocate=False,
        )

    def compute_voltage(self, sim_data: SimulationData) -> FreqDataArray:
        """Helper to compute voltage across the port."""
        field_data = sim_data[self._voltage_monitor_name]
        size = list(self.size)
        size[self.current_axis] = 0
        voltage_integral = VoltageIntegralAxisAligned(
            center=self.center,
            size=size,
            extrapolate_to_endpoints=True,
            snap_path_to_grid=True,
            sign="+",
        )
        voltage = voltage_integral.compute_voltage(field_data)
        # Return data array of voltage with coordinates of frequency
        return voltage

    def compute_current(self, sim_data: SimulationData) -> FreqDataArray:
        """Helper to compute current flowing through the port."""
        # Diagram of contour integral, dashed line indicates location of sheet resistance
        # and electric field used for voltage computation. Voltage axis is out-of-page.
        #
        #                                    current_axis = ->
        #                                    injection_axis = ^
        #
        #                  |                   h2_field ->             |
        #    h_cap_minus ^  -------------------------------------------  h_cap_plus ^
        #                  |                   h1_field ->             |

        field_data = sim_data[self._current_monitor_name]
        # Get h field tangent to resistive sheet
        h_component = "xyz"[self.current_axis]
        inject_component = "xyz"[self.injection_axis]
        # monitor_data = sim_data[f"{port.name}_H{h_component}"]
        field_components = field_data.field_components
        h_field = field_components[f"H{h_component}"]
        # Coordinates as numpy array for h_field along curren and injection axis
        h_coords_along_current = h_field.coords[h_component].values
        h_coords_along_injection = h_field.coords[inject_component].values
        # h_cap represents the very short section (single cell) of the H contour that
        # is in the injection_axis direction. It is needed to fully enclose the sheet.
        h_cap_field = field_components[f"H{inject_component}"]
        # Coordinates of h_cap field as numpy arrays
        h_cap_coords_along_current = h_cap_field.coords[h_component].values
        h_cap_coords_along_injection = h_cap_field.coords[inject_component].values

        # Use the coordinates of h_cap since it lies on the same grid that the
        # lumped resistor is snapped to
        orth_index = np.argmin(
            np.abs(h_cap_coords_along_injection - self.center[self.injection_axis])
        )
        inject_center = h_cap_coords_along_injection[orth_index]
        # Some sanity checks, tangent H field coordinates should be directly above
        # and below the coordinates of the resistive sheet
        error_message = (
            "Unexpected error encountered when setting up the current computation for a 'LumpedPort'. "
            "If you encounter this error, please create an issue in the Tidy3D github repository."
        )
        if orth_index <= 0:
            raise AssertionError(error_message)
        if inject_center >= h_coords_along_injection[orth_index]:
            raise AssertionError(error_message)
        if h_coords_along_injection[orth_index - 1] >= inject_center:
            raise AssertionError(error_message)
        # Distance between the h1_field and h2_field, a single cell size
        dcap = h_coords_along_injection[orth_index] - h_coords_along_injection[orth_index - 1]

        # Next find the size in the current_axis direction
        # Find exact bounds of port taking into consideration the Yee grid
        # Select bounds carefully and allow for h_cap very close to the port bounds
        port_min = self.bounds[0][self.current_axis]
        port_max = self.bounds[1][self.current_axis]

        (idx_min, idx_max) = LumpedPort._select_within_bounds(
            h_coords_along_current, port_min, port_max
        )
        # Use these indices to select the exact positions of the h_cap field
        h_min_bound = h_cap_coords_along_current[idx_min - 1]
        h_max_bound = h_cap_coords_along_current[idx_max]

        # Setup axis aligned contour integral, which is defined by a plane
        # The path integral is snapped to the grid, so center and size will
        # be slightly modified when compared to the original port.
        center = list(self.center)
        center[self.injection_axis] = inject_center
        center[self.current_axis] = (h_max_bound + h_min_bound) / 2
        size = [0, 0, 0]
        size[self.current_axis] = h_max_bound - h_min_bound
        size[self.injection_axis] = dcap

        # H field is continuous at integral bounds, so extrapolation is turned off
        I_integral = CurrentIntegralAxisAligned(
            center=center,
            size=size,
            sign="+",
            extrapolate_to_endpoints=False,
            snap_contour_to_grid=True,
        )
        return I_integral.compute_current(field_data)

    @staticmethod
    def _select_within_bounds(coords: np.array, min, max):
        """Helper to return indices of coordinates within min and max bounds,
        including a tolerance. xarray does not have this functionality yet.
        """
        min_idx = np.searchsorted(coords, min, "left")
        # If a coordinate is close enough, it is considered included
        if min_idx > 0 and np.isclose(coords[min_idx - 1], min, rtol=fp_eps, atol=fp_eps):
            min_idx -= 1
        max_idx = np.searchsorted(coords, max, "left")
        if max_idx < len(coords) and np.isclose(coords[max_idx], max, rtol=fp_eps, atol=fp_eps):
            max_idx += 1

        return (min_idx, max_idx - 1)

    def _check_grid_size(self, yee_grid: YeeGrid):
        """Raises :class:``SetupError`` if the grid is too coarse at port locations."""
        e_component = "xyz"[self.voltage_axis]
        e_yee_grid = yee_grid.grid_dict[f"E{e_component}"]
        coords = e_yee_grid.to_dict[e_component]
        min_bound = self.bounds[0][self.voltage_axis]
        max_bound = self.bounds[1][self.voltage_axis]
        coords_within_port = np.any(np.logical_and(coords > min_bound, coords < max_bound))
        if not coords_within_port:
            raise SetupError(
                f"Grid is too coarse along '{e_component}' direction for the lumped port "
                f"at location '{self.center}'. Either set the port's 'num_grid_cells' to "
                f"a nonzero integer or modify the 'GridSpec'."
            )



"""Defines heat simulation data class"""

from __future__ import annotations

from typing import Tuple

import numpy as np
import pydantic.v1 as pd

from ....exceptions import DataError
from ...base_sim.data.sim_data import AbstractSimulationData
from ...data.data_array import SpatialDataArray
from ...data.dataset import TetrahedralGridDataset, TriangularGridDataset, UnstructuredGridDataset
from ...types import Ax, Literal, RealFieldVal
from ...viz import add_ax_if_none, equal_aspect
from ..simulation import HeatSimulation
from .monitor_data import HeatMonitorDataType, TemperatureData


class HeatSimulationData(AbstractSimulationData):
    """Stores results of a heat simulation.

    Example
    -------
    >>> from tidy3d import Medium, SolidSpec, FluidSpec, UniformUnstructuredGrid, SpatialDataArray
    >>> from tidy3d import Structure, Box, UniformUnstructuredGrid, UniformHeatSource
    >>> from tidy3d import StructureBoundary, TemperatureBC, TemperatureMonitor, TemperatureData
    >>> from tidy3d import HeatBoundarySpec
    >>> import numpy as np
    >>> temp_mnt = TemperatureMonitor(size=(1, 2, 3), name="sample")
    >>> heat_sim = HeatSimulation(
    ...     size=(3.0, 3.0, 3.0),
    ...     structures=[
    ...         Structure(
    ...             geometry=Box(size=(1, 1, 1), center=(0, 0, 0)),
    ...             medium=Medium(
    ...                 permittivity=2.0, heat_spec=SolidSpec(
    ...                     conductivity=1,
    ...                     capacity=1,
    ...                 )
    ...             ),
    ...             name="box",
    ...         ),
    ...     ],
    ...     medium=Medium(permittivity=3.0, heat_spec=FluidSpec()),
    ...     grid_spec=UniformUnstructuredGrid(dl=0.1),
    ...     sources=[UniformHeatSource(rate=1, structures=["box"])],
    ...     boundary_spec=[
    ...         HeatBoundarySpec(
    ...             placement=StructureBoundary(structure="box"),
    ...             condition=TemperatureBC(temperature=500),
    ...         )
    ...     ],
    ...     monitors=[temp_mnt],
    ... )
    >>> x = [1,2]
    >>> y = [2,3,4]
    >>> z = [3,4,5,6]
    >>> coords = dict(x=x, y=y, z=z)
    >>> temp_array = SpatialDataArray(300 * np.abs(np.random.random((2,3,4))), coords=coords)
    >>> temp_mnt_data = TemperatureData(monitor=temp_mnt, temperature=temp_array)
    >>> heat_sim_data = HeatSimulationData(
    ...     simulation=heat_sim, data=[temp_mnt_data],
    ... )
    """

    simulation: HeatSimulation = pd.Field(
        title="Heat Simulation",
        description="Original :class:`.HeatSimulation` associated with the data.",
    )

    data: Tuple[HeatMonitorDataType, ...] = pd.Field(
        ...,
        title="Monitor Data",
        description="List of :class:`.MonitorData` instances "
        "associated with the monitors of the original :class:`.Simulation`.",
    )

    @equal_aspect
    @add_ax_if_none
    def plot_field(
        self,
        monitor_name: str,
        val: RealFieldVal = "real",
        scale: Literal["lin", "log"] = "lin",
        structures_alpha: float = 0.2,
        robust: bool = True,
        vmin: float = None,
        vmax: float = None,
        ax: Ax = None,
        **sel_kwargs,
    ) -> Ax:
        """Plot the data for a monitor with simulation plot overlaid.

        Parameters
        ----------
        field_monitor_name : str
            Name of :class:`.TemperatureMonitorData` to plot.
        val : Literal['real', 'abs', 'abs^2'] = 'real'
            Which part of the field to plot.
        scale : Literal['lin', 'log']
            Plot in linear or logarithmic scale.
        structures_alpha : float = 0.2
            Opacity of the structure permittivity.
            Must be between 0 and 1 (inclusive).
        robust : bool = True
            If True and vmin or vmax are absent, uses the 2nd and 98th percentiles of the data
            to compute the color limits. This helps in visualizing the field patterns especially
            in the presence of a source.
        vmin : float = None
            The lower bound of data range that the colormap covers. If ``None``, they are
            inferred from the data and other keyword arguments.
        vmax : float = None
            The upper bound of data range that the colormap covers. If ``None``, they are
            inferred from the data and other keyword arguments.
        ax : matplotlib.axes._subplots.Axes = None
            matplotlib axes to plot on, if not specified, one is created.
        sel_kwargs : keyword arguments used to perform ``.sel()`` selection in the monitor data.
            These kwargs can select over the spatial dimensions (``x``, ``y``, ``z``),
            or time dimension (``t``) if applicable.
            For the plotting to work appropriately, the resulting data after selection must contain
            only two coordinates with len > 1.
            Furthermore, these should be spatial coordinates (``x``, ``y``, or ``z``).

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """

        monitor_data = self[monitor_name]

        if not isinstance(monitor_data, TemperatureData):
            raise DataError(
                f"Monitor '{monitor_name}' (type '{monitor_data.monitor.type}') is not a "
                f"'TemperatureMonitor'."
            )

        if monitor_data.temperature is None:
            raise DataError(f"No data to plot for monitor '{monitor_name}'.")

        field_data = self._field_component_value(monitor_data.temperature, val)

        if val == "abs^2":
            field_name = "|T|², K²"
        else:
            field_name = "T, K"

        if scale == "log":
            field_data = np.log10(np.abs(field_data))

        cmap = "coolwarm"

        # do sel on unstructured data
        # it could produce either SpatialDataArray or UnstructuredGridDatasetType
        if isinstance(field_data, UnstructuredGridDataset) and len(sel_kwargs) > 0:
            field_data = field_data.sel(**sel_kwargs)

        if isinstance(field_data, TetrahedralGridDataset):
            raise DataError(
                "Must select a two-dimensional slice of unstructured dataset for plotting"
                " on a plane."
            )

        if isinstance(field_data, TriangularGridDataset):
            field_data.plot(
                ax=ax,
                cmap=cmap,
                vmin=vmin,
                vmax=vmax,
                cbar_kwargs={"label": field_name},
                grid=False,
            )

            # compute parameters for structures overlay plot
            axis = field_data.normal_axis
            position = field_data.normal_pos

            # compute plot bounds
            field_data_bounds = field_data.bounds
            min_bounds = list(field_data_bounds[0])
            max_bounds = list(field_data_bounds[1])
            min_bounds.pop(axis)
            max_bounds.pop(axis)

        if isinstance(field_data, SpatialDataArray):
            # interp out any monitor.size==0 dimensions
            monitor = self.simulation.get_monitor_by_name(monitor_name)
            thin_dims = {
                "xyz"[dim]: monitor.center[dim]
                for dim in range(3)
                if monitor.size[dim] == 0 and "xyz"[dim] not in sel_kwargs
            }
            for axis, pos in thin_dims.items():
                if field_data.coords[axis].size <= 1:
                    field_data = field_data.sel(**{axis: pos}, method="nearest")
                else:
                    field_data = field_data.interp(**{axis: pos}, kwargs=dict(bounds_error=True))

            # select the extra coordinates out of the data from user-specified kwargs
            for coord_name, coord_val in sel_kwargs.items():
                if field_data.coords[coord_name].size <= 1:
                    field_data = field_data.sel(**{coord_name: coord_val}, method=None)
                else:
                    field_data = field_data.interp(
                        **{coord_name: coord_val}, kwargs=dict(bounds_error=True)
                    )

            field_data = field_data.squeeze(drop=True)
            non_scalar_coords = {name: c for name, c in field_data.coords.items() if c.size > 1}

            # assert the data is valid for plotting
            if len(non_scalar_coords) != 2:
                raise DataError(
                    f"Data after selection has {len(non_scalar_coords)} coordinates "
                    f"({list(non_scalar_coords.keys())}), "
                    "must be 2 spatial coordinates for plotting on plane. "
                    "Please add keyword arguments to 'plot_monitor_data()' to select out the other coords."
                )

            spatial_coords_in_data = {
                coord_name: (coord_name in non_scalar_coords) for coord_name in "xyz"
            }

            if sum(spatial_coords_in_data.values()) != 2:
                raise DataError(
                    "All coordinates in the data after selection must be spatial (x, y, z), "
                    f" given {non_scalar_coords.keys()}."
                )

            # get the spatial coordinate corresponding to the plane
            planar_coord = [name for name, c in spatial_coords_in_data.items() if c is False][0]
            axis = "xyz".index(planar_coord)
            position = float(field_data.coords[planar_coord])

            xy_coord_labels = list("xyz")
            xy_coord_labels.pop(axis)
            x_coord_label, y_coord_label = xy_coord_labels[0], xy_coord_labels[1]
            field_data.plot(
                ax=ax,
                x=x_coord_label,
                y=y_coord_label,
                cmap=cmap,
                vmin=vmin,
                vmax=vmax,
                robust=robust,
                cbar_kwargs={"label": field_name},
            )

            # compute plot bounds
            x_coord_values = field_data.coords[x_coord_label]
            y_coord_values = field_data.coords[y_coord_label]
            min_bounds = (min(x_coord_values), min(y_coord_values))
            max_bounds = (max(x_coord_values), max(y_coord_values))

        # select the cross section data
        interp_kwarg = {"xyz"[axis]: position}
        # plot the simulation heat conductivity
        ax = self.simulation.scene.plot_structures_heat_conductivity(
            cbar=False,
            alpha=structures_alpha,
            ax=ax,
            **interp_kwarg,
        )

        # set the limits based on the xarray coordinates min and max
        ax.set_xlim(min_bounds[0], max_bounds[0])
        ax.set_ylim(min_bounds[1], max_bounds[1])

        return ax



"""Defines jax-compatible DataArrays."""

from __future__ import annotations

from typing import Any, Dict, List, Literal, Sequence, Tuple, Union

import h5py
import jax
import jax.numpy as jnp
import numpy as np
import pydantic.v1 as pd
import xarray as xr
from jax.tree_util import register_pytree_node_class

from .....components.base import Tidy3dBaseModel, cached_property, skip_if_fields_missing
from .....exceptions import AdjointError, DataError, Tidy3dKeyError

# condition setting when to set value in DataArray to zero:
# if abs(val) <= VALUE_FILTER_THRESHOLD * max(abs(val))
VALUE_FILTER_THRESHOLD = 1e-6

# JaxDataArrays are written to json as JAX_DATA_ARRAY_TAG
JAX_DATA_ARRAY_TAG = "<<JaxDataArray>>"


@register_pytree_node_class
class JaxDataArray(Tidy3dBaseModel):
    """A :class:`.DataArray`-like class that only wraps xarray for jax compatibility."""

    values: Any = pd.Field(
        ...,
        title="Values",
        description="Nested list containing the raw values, which can be tracked by jax.",
        jax_field=True,
    )

    coords: Dict[str, list] = pd.Field(
        ...,
        title="Coords",
        description="Dictionary storing the coordinates, namely ``(direction, f, mode_index)``.",
    )

    def to_tidy3d(self: JaxDataArray) -> xr.DataArray:
        """Convert :class:`.JaxDataArray` instance to ``xr.DataArray`` instance."""
        coords = {k: np.array(v).tolist() for k, v in self.coords.items()}
        return xr.DataArray(np.array(self.values), coords=coords, dims=self.coords.keys())

    @classmethod
    def from_tidy3d(cls, tidy3d_obj: xr.DataArray) -> JaxDataArray:
        """Convert ``xr.DataArray`` instance to :class:`.JaxDataArray`."""
        coords = {k: np.array(v).tolist() for k, v in tidy3d_obj.coords.items()}
        return cls(values=tidy3d_obj.data, coords=coords)

    @pd.validator("values", always=True)
    def _convert_values_to_np(cls, val):
        """Convert supplied values to numpy if they are list (from file)."""
        if isinstance(val, list):
            return np.array(val)
        return val

    @pd.validator("coords", always=True)
    @skip_if_fields_missing(["values"])
    def _coords_match_values(cls, val, values):
        """Make sure the coordinate dimensions and shapes match the values data."""

        _values = values.get("values")

        # get the shape, handling both regular and jax objects
        try:
            values_shape = np.array(_values).shape
        except TypeError:
            values_shape = jnp.array(_values).shape

        for (key, coord_val), size_dim in zip(val.items(), values_shape):
            if len(coord_val) != size_dim:
                raise ValueError(
                    f"JaxDataArray coord {key} has {len(coord_val)} elements, "
                    "which doesn't match the values array "
                    f"with size {size_dim} along that dimension."
                )

        return val

    @pd.validator("coords", always=True)
    def _convert_coords_to_list(cls, val):
        """Convert supplied coordinates to Dict[str, list]."""
        return {coord_name: list(coord_list) for coord_name, coord_list in val.items()}

    def __eq__(self, other) -> bool:
        """Check if two ``JaxDataArray`` instances are equal."""
        return jnp.array_equal(self.values, other.values)

    def to_hdf5(self, fname: str, group_path: str) -> None:
        """Save an xr.DataArray to the hdf5 file with a given path to the group."""
        sub_group = fname.create_group(group_path)
        sub_group["values"] = self.values
        dims = []
        for key, val in self.coords.items():
            # sub_group[key] = val
            dims.append(key)
            val = np.array(val)
            if val.dtype == "<U1":
                sub_group[key] = val.tolist()
            else:
                sub_group[key] = val
        sub_group["dims"] = dims

    @classmethod
    def from_hdf5(cls, fname: str, group_path: str) -> JaxDataArray:
        """Load an DataArray from an hdf5 file with a given path to the group."""
        with h5py.File(fname, "r") as f:
            sub_group = f[group_path]
            values = np.array(sub_group["values"])
            dims = sub_group["dims"]
            coords = {dim: np.array(sub_group[dim]) for dim in dims}
            for key, val in coords.items():
                val = np.array(val)
                if val.dtype == "O":
                    coords[key] = [byte_string.decode() for byte_string in val.tolist()]

            coords = {
                key: val.tolist() if isinstance(val, np.ndarray) else val
                for key, val in coords.items()
            }
            return cls(values=values, coords=coords)

    @cached_property
    def as_ndarray(self) -> np.ndarray:
        """``self.values`` as a numpy array."""
        if not isinstance(self.values, np.ndarray):
            return np.array(self.values)
        return self.values

    @cached_property
    def as_jnp_array(self) -> jnp.ndarray:
        """``self.values`` as a jax array."""
        if not isinstance(self.values, jnp.ndarray):
            return jnp.array(self.values)
        return self.values

    @cached_property
    def shape(self) -> tuple:
        """Shape of self.values."""
        return self.as_jnp_array.shape

    @cached_property
    def as_list(self) -> list:
        """``self.values`` as a numpy array converted to a list."""
        return self.as_ndarray.tolist()

    @cached_property
    def real(self) -> np.ndarray:
        """Real part of self."""
        new_values = jnp.real(self.as_jnp_array)
        return self.copy(update=dict(values=new_values))

    @cached_property
    def imag(self) -> np.ndarray:
        """Imaginary part of self."""
        new_values = jnp.imag(self.as_jnp_array)
        return self.copy(update=dict(values=new_values))

    def conj(self) -> JaxDataArray:
        """Complex conjugate of self."""
        new_values = jnp.conj(self.as_jnp_array)
        return self.copy(update=dict(values=new_values))

    def __abs__(self) -> JaxDataArray:
        """Absolute value of self's values."""
        new_values = jnp.abs(self.as_jnp_array)
        return self.updated_copy(values=new_values)

    def __pow__(self, power: int) -> JaxDataArray:
        """Values raised to a power."""
        new_values = self.as_jnp_array**power
        return self.updated_copy(values=new_values)

    def __add__(self, other: JaxDataArray) -> JaxDataArray:
        """Sum self with something else."""
        if isinstance(other, JaxDataArray):
            new_values = self.as_jnp_array + other.as_jnp_array
        else:
            new_values = self.as_jnp_array + other
        return self.updated_copy(values=new_values)

    def __neg__(self) -> JaxDataArray:
        """Negative of self."""
        new_values = -self.as_jnp_array
        return self.updated_copy(values=new_values)

    def __sub__(self, other) -> JaxDataArray:
        """Subtraction"""
        return self + (-other)

    def __radd__(self, other) -> JaxDataArray:
        """Sum self with something else."""
        return self + other

    def __mul__(self, other: JaxDataArray) -> JaxDataArray:
        """Multiply self with something else."""
        if isinstance(other, JaxDataArray):
            new_values = self.as_jnp_array * other.as_jnp_array
        elif isinstance(other, xr.DataArray):
            # handle case where other is missing dims present in self
            new_shape = list(self.shape)
            for dim_index, dim in enumerate(self.coords.keys()):
                if dim not in other.dims:
                    other = other.expand_dims(dim=dim)
                    new_shape[dim_index] = 1

            other_values = other.values.reshape(new_shape)
            new_values = self.as_jnp_array * other_values
        else:
            new_values = self.as_jnp_array * other

        return self.updated_copy(values=new_values)

    def __rmul__(self, other) -> JaxDataArray:
        """Multiply self with something else."""
        return self * other

    def sum(self, dim: str = None):
        """Sum (optionally along a single or multiple dimensions)."""

        if dim is None:
            return jnp.sum(self.as_jnp_array)

        # dim is supplied
        if isinstance(dim, str):
            axis = list(self.coords.keys()).index(dim)
            new_values = jnp.sum(self.values, axis=axis)
            new_coords = self.coords.copy()
            new_coords.pop(dim)
            return self.updated_copy(values=new_values, coords=new_coords)

        # dim is iterative, recursively call sum with single dim
        ret = self.copy()
        for dim_i in dim:
            ret = ret.sum(dim=dim_i)
        return ret

    def squeeze(self, dim: str = None, drop: bool = True) -> JaxDataArray:
        """Remove any non-zero dims."""

        if dim is None:
            new_values = jnp.squeeze(self.as_jnp_array)

            new_coords = {}
            for (key, val), dim_size in zip(self.coords.items(), self.values.shape):
                if dim_size > 1:
                    new_coords.update({key: val})

        else:
            axis = list(self.coords.keys()).index(dim)
            new_values = jnp.array(jnp.squeeze(self.as_jnp_array, axis=axis))
            new_coords = self.coords.copy()
            new_coords.pop(dim)
        return self.updated_copy(values=new_values, coords=new_coords)

    def get_coord_list(self, coord_name: str) -> list:
        """Get a coordinate list by name."""

        if coord_name not in self.coords:
            raise Tidy3dKeyError(f"Could not select '{coord_name}', not found in coords dict.")
        return self.coords.get(coord_name)

    def isel_single(self, coord_name: str, coord_index: int) -> JaxDataArray:
        """Select a value corresponding to a single coordinate from the :class:`.JaxDataArray`."""

        # select out the proper values and coordinates
        coord_axis = list(self.coords.keys()).index(coord_name)
        values = self.as_jnp_array
        new_values = jnp.take(values, indices=coord_index, axis=coord_axis)
        new_coords = self.coords.copy()

        # if the coord index has more than one item, keep that coordinate
        coord_index = np.array(coord_index)
        if len(coord_index.shape) >= 1:
            coord_indices = coord_index.tolist()
            new_coord_vals = [self.coords[coord_name][coord_index] for coord_index in coord_indices]
            new_coords[coord_name] = new_coord_vals
        else:
            new_coords.pop(coord_name)

        # return just the values if no coordinate remain
        if not new_coords:
            if new_values.shape:
                raise AdjointError(
                    "All coordinates selected out, but raw data values are still multi-dimensional."
                    " If you encountered this error, please raise an issue on the Tidy3D "
                    "front end github repository so we can look into the source of the bug."
                )

            return new_values

        # otherwise, return another JaxDataArray with the values and coords selected out
        return self.copy(update=dict(values=new_values, coords=new_coords))

    def isel(self, **isel_kwargs) -> JaxDataArray:
        """Select a value from the :class:`.JaxDataArray` by indexing into coordinates by index."""

        self_sel = self.copy()
        for coord_name, coord_index in isel_kwargs.items():
            coord_index = np.array(coord_index)
            coord_list = self_sel.get_coord_list(coord_name)
            if np.any(coord_index < 0) or np.any(coord_index >= len(coord_list)):
                raise DataError(
                    f"'isel' kwarg '{coord_name}={coord_index}' is out of range "
                    f"for the coordinate '{coord_name}' with {len(coord_list)} values."
                )
            self_sel = self_sel.isel_single(coord_name=coord_name, coord_index=coord_index)

        return self_sel

    def sel(
        self, indexers: dict = None, method: Literal[None, "nearest"] = None, **sel_kwargs
    ) -> JaxDataArray:
        """Select a value from the :class:`.JaxDataArray` by indexing into coordinates by value.

        Parameters
        ----------
        sel_kwargs : dict
            Keyword arguments with names matching the coordinates of :class:`.JaxDataArray` and values
            given by scalars or lists, e.g. `da.sel(x=0.1, y=[0.2, 0.3])`.
        method : Literal[None, "nearest"] = None
            Method to use for matching coordinate values:

            - None (default): only exact matches
            - nearest: use nearest valid index value

        Returns
        -------
        JaxDataArray
            JaxDataArray with extracted values.
        """
        if method not in [None, "nearest"]:
            raise NotImplementedError(f"Unkown selection method: {method}.")

        isel_kwargs = {}
        for coord_name, vals in sel_kwargs.items():
            coord_list = self.get_coord_list(coord_name)

            try:  # handle non-numeric types (e.g. str)
                coord_list = jnp.asarray(coord_list)
            except TypeError:
                isel_kwargs[coord_name] = self._indices_literal(coord_list, vals)
                continue

            vals_ary = jnp.atleast_1d(vals)
            dist = jnp.abs(coord_list[None] - vals_ary[:, None])

            if method is None:
                indices = jnp.where(jnp.isclose(dist, 0))[1]
            elif method == "nearest":
                indices = jnp.argmin(dist, axis=1)

            if indices.size == 0:
                raise DataError(
                    f"Could not select '{coord_name}={vals_ary}', some values were not found."
                )

            if np.isscalar(vals):
                indices = jnp.squeeze(indices)

            isel_kwargs[coord_name] = indices

        return self.isel(**isel_kwargs)

    def _indices_literal(self, coord_list: list, values: Union[Any, Sequence[Any]]) -> np.ndarray:
        """Find indices of non-numeric `values` in `coord_list`.

        Parameters
        ----------
        coord_list : list
            List of all entries for a specific coordinate.
        values : Union[Any, Sequence[Any]]
            Single value or values of which to find the index of.

        Returns
        -------
        numpy.ndarray
            Indices of `values` in `coord_list`.
        """
        indices = []
        for v in np.atleast_1d(values):
            if v not in coord_list:
                raise DataError(f"Could not select '{v}' from '{coord_list}', value not found.")
            indices.append(coord_list.index(v))

        if np.isscalar(values):
            indices = np.squeeze(indices)

        return indices

    def assign_coords(self, coords: dict = None, **coords_kwargs) -> JaxDataArray:
        """Assign new coordinates to this object."""

        update_kwargs = self.coords.copy()

        for key, val in coords_kwargs.items():
            update_kwargs[key] = val

        if coords:
            for key, val in coords.items():
                update_kwargs[key] = val

        update_kwargs = {key: np.array(value).tolist() for key, value in update_kwargs.items()}
        return self.updated_copy(coords=update_kwargs)

    def multiply_at(self, value: complex, coord_name: str, indices: List[int]) -> JaxDataArray:
        """Multiply self by value at indices into ."""
        axis = list(self.coords.keys()).index(coord_name)
        scalar_data_arr = self.as_jnp_array
        scalar_data_arr = jnp.moveaxis(scalar_data_arr, axis, 0)
        scalar_data_arr = scalar_data_arr.at[indices].multiply(value)
        scalar_data_arr = jnp.moveaxis(scalar_data_arr, 0, axis)
        return self.updated_copy(values=scalar_data_arr)

    def interp_single(self, key: str, val: float) -> JaxDataArray:
        """Interpolate into a single dimension of self.

        Note: this interpolation works by finding the index of the value into the coords list.
        Instead of an integer value, we use interpolation to get a floating point index.
        The floor() of this value is the 'minus' index and the ceil() gives the 'plus' index.
        We then apply coefficients linearly based on how close to `plus` or minus we are.
        This is a workaround to `jnp.interp` not allowing multi-dimensional interpolation.
        """

        val = jax.lax.stop_gradient(val)

        # get the coordinates associated with this key.
        if key not in self.coords:
            raise Tidy3dKeyError(f"Key '{key}' not found in JaxDataArray coords.")
        coords_1d = jnp.array(self.coords[key])
        axis = list(self.coords.keys()).index(key)

        # get floating point index of the value into these coordinates
        coord_indices = jnp.arange(len(coords_1d))
        index_interp = jnp.interp(x=val, xp=coords_1d, fp=coord_indices)

        # strip out the linear interpolation coefficients from the float index
        index_minus = np.array(index_interp).astype(int)
        index_plus = index_minus + 1
        coeff_plus = index_interp - index_minus

        # if any plus_index is out of range, set it in range (coeff will be 0 anyway)
        if index_plus.shape:
            index_plus[index_plus >= len(coord_indices)] = len(coord_indices) - 1
        else:
            if index_plus > len(coord_indices):
                index_plus = index_minus
                coeff_plus = 0.0

        coeff_minus = 1 - coeff_plus

        def get_values_at_index(key: str, index: int) -> jnp.array:
            """grab values array at index into coordinate key."""
            values_sel = self.isel(**{key: index})
            if isinstance(values_sel, JaxDataArray):
                return values_sel.values
            return values_sel

        # return weighted average of this object along these dimensions
        values_minus = get_values_at_index(key=key, index=index_minus)

        if np.any(coeff_plus > 0):
            values_plus = get_values_at_index(key=key, index=index_plus)

            if coeff_minus.shape:
                coeff_shape = np.ones(len(values_minus.shape), dtype=int)
                coeff_shape[axis] = len(coeff_minus)
                coeff_minus = coeff_minus.reshape(coeff_shape)
                coeff_plus = coeff_plus.reshape(coeff_shape)

            values_interp = coeff_minus * values_minus + coeff_plus * values_plus
        else:
            values_interp = values_minus

        # construct a new JaxDataArray to return
        coords_interp = self.coords.copy()
        if jnp.array(index_interp).size <= 1:
            coords_interp.pop(key)
        else:
            coords_interp[key] = np.atleast_1d(val).tolist()

        if coords_interp:
            return JaxDataArray(values=values_interp, coords=coords_interp)
        return values_interp

    def interp(self, kwargs=None, assume_sorted=None, **interp_kwargs) -> JaxDataArray:
        """Linearly interpolate into the :class:`.JaxDataArray` at values into coordinates."""

        # note: kwargs does nothing, only used for making this subclass compatible with super

        ret_value = self.copy()
        for key, val in interp_kwargs.items():
            ret_value = ret_value.interp_single(key=key, val=val)
        return ret_value

    @cached_property
    def nonzero_val_coords(self) -> Tuple[List[complex], Dict[str, Any]]:
        """The value and coordinate associated with the only non-zero element of ``self.values``."""

        values = np.nan_to_num(self.as_ndarray)

        # filter out values that are very small relative to maximum
        values_filtered = values.copy()
        max_value = np.max(np.abs(values_filtered))
        val_cutoff = VALUE_FILTER_THRESHOLD * max_value
        values_filtered[np.abs(values_filtered) <= val_cutoff] = 0.0

        nonzero_inds = np.nonzero(values_filtered)
        nonzero_values = values_filtered[nonzero_inds].tolist()

        nonzero_coords = {}
        for nz_inds, (coord_name, coord_list) in zip(nonzero_inds, self.coords.items()):
            coord_array = np.array(coord_list)
            nonzero_coords[coord_name] = coord_array[nz_inds].tolist()

        return nonzero_values, nonzero_coords

    def tree_flatten(self) -> Tuple[list, dict]:
        """Jax works on the values, stash the coords for reconstruction."""

        return self.values, self.coords

    @classmethod
    def tree_unflatten(cls, aux_data, children) -> JaxDataArray:
        """How to unflatten the values and coords."""

        return cls(values=children, coords=aux_data)



import itertools

import autograd.numpy as anp
import numpy as np
from numpy.typing import NDArray
from scipy.interpolate import RegularGridInterpolator

from .types import InterpolationType


def _evaluate_nearest(
    indices: NDArray[np.int64], norm_distances: NDArray[np.float64], values: NDArray[np.float64]
) -> NDArray[np.float64]:
    """Perform nearest neighbor interpolation in an n-dimensional space.

    This function determines the nearest neighbor in a grid for a given point
    and returns the corresponding value from the input array.

    Parameters
    ----------
    indices : NDArray[np.int64]
        Indices of the lower bounds of the grid cell containing the interpolation point.
    norm_distances : NDArray[np.float64]
        Normalized distances from the lower bounds of the grid cell to the
        interpolation point, for each dimension.
    values : NDArray[np.float64]
        The n-dimensional array of values to interpolate from.

    Returns
    -------
    NDArray[np.float64]
        The value of the nearest neighbor to the interpolation point.
    """
    idx_res = tuple(anp.where(yi <= 0.5, i, i + 1) for i, yi in zip(indices, norm_distances))
    return values[idx_res]


def _evaluate_linear(
    indices: NDArray[np.int64], norm_distances: NDArray[np.float64], values: NDArray[np.float64]
) -> NDArray[np.float64]:
    """Perform linear interpolation in an n-dimensional space.

    This function calculates the linearly interpolated value at a point in an
    n-dimensional grid, given the indices of the surrounding grid points and
    the normalized distances to these points.
    The multi-linear interpolation is implemented by computing a weighted
    average of the values at the vertices of the hypercube surrounding the
    interpolation point.

    Parameters
    ----------
    indices : NDArray[np.int64]
        Indices of the lower bounds of the grid cell containing the interpolation point.
    norm_distances : NDArray[np.float64]
        Normalized distances from the lower bounds of the grid cell to the
        interpolation point, for each dimension.
    values : NDArray[np.float64]
        The n-dimensional array of values to interpolate from.

    Returns
    -------
    NDArray[np.float64]
        The interpolated value at the desired point.
    """
    # Create a slice object for broadcasting over trailing dimensions
    _slice = (slice(None),) + (None,) * (values.ndim - len(indices))

    # Prepare iterables for lower and upper bounds of the hypercube
    ix = zip(indices, (1 - yi for yi in norm_distances))
    iy = zip((i + 1 for i in indices), norm_distances)

    # Initialize the result
    value = anp.zeros(1)

    # Iterate over all vertices of the hypercube
    for h in itertools.product(*zip(ix, iy)):
        edge_indices, weights = zip(*h)

        # Compute the weight for this vertex
        weight = anp.ones(1)
        for w in weights:
            weight = weight * w

        # Compute the contribution of this vertex and add it to the result
        term = values[edge_indices] * weight[_slice]
        value = value + term

    return value


def interpn(
    points: tuple[NDArray[np.float64], ...],
    values: NDArray[np.float64],
    xi: tuple[NDArray[np.float64], ...],
    *,
    method: InterpolationType = "linear",
) -> NDArray[np.float64]:
    """Interpolate over a rectilinear grid in arbitrary dimensions.

    This function mirrors the interface of `scipy.interpolate.interpn` but is differentiable with autograd.

    Parameters
    ----------
    points : tuple[NDArray[np.float64], ...]
        The points defining the rectilinear grid in n dimensions.
    values : NDArray[np.float64]
        The data values on the rectilinear grid.
    xi : tuple[NDArray[np.float64], ...]
        The coordinates to sample the gridded data at.
    method : InterpolationType = "linear"
        The method of interpolation to perform. Supported are "linear" and "nearest".

    Returns
    -------
    NDArray[np.float64]
        The interpolated values.

    Raises
    ------
    ValueError
        If the interpolation method is not supported.

    See Also
    --------
    `scipy.interpolate.interpn <https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.interpn.html>`_
    """
    if method == "nearest":
        interp_fn = _evaluate_nearest
    elif method == "linear":
        interp_fn = _evaluate_linear
    else:
        raise ValueError(f"Unsupported interpolation method: {method}")

    itrp = RegularGridInterpolator(points, values, method=method)
    grid = anp.meshgrid(*xi, indexing="ij")

    # Prepare the grid for interpolation
    # This step reshapes the grid, checks for NaNs and out-of-bounds values
    # It returns:
    #   - reshaped grid
    #   - original shape
    #   - number of dimensions
    #   - boolean array indicating NaN positions
    #   - (discarded) boolean array for out-of-bounds values
    grid, shape, ndim, nans, _ = itrp._prepare_xi(tuple(grid))

    # Find the indices of the grid cells containing the interpolation points
    # and calculate the normalized distances (ranging from 0 at lower grid point to 1
    # at upper grid point) within these cells
    indices, norm_distances = itrp._find_indices(grid.T)

    result = interp_fn(indices, norm_distances, values)
    nans = anp.reshape(nans, (-1,) + (1,) * (result.ndim - 1))
    result = anp.where(nans, np.nan, result)
    return anp.reshape(result, shape[:-1] + values.shape[ndim:])


__all__ = [
    "interpn",
]



"""Rectangular dielectric waveguide utilities."""

from typing import Any, List, Tuple, Union

import numpy
import pydantic.v1 as pydantic
from matplotlib import pyplot
from typing_extensions import Annotated

from ...components.base import Tidy3dBaseModel, cached_property, skip_if_fields_missing
from ...components.boundary import BoundarySpec, Periodic
from ...components.data.data_array import FreqModeDataArray, ModeIndexDataArray
from ...components.geometry.base import Box
from ...components.geometry.polyslab import PolySlab
from ...components.grid.grid_spec import GridSpec
from ...components.medium import Medium, MediumType
from ...components.mode import ModeSpec
from ...components.simulation import Simulation
from ...components.source import GaussianPulse, ModeSource
from ...components.structure import Structure
from ...components.types import TYPE_TAG_STR, ArrayFloat1D, Ax, Axis, Coordinate, Literal, Size1D
from ...components.viz import add_ax_if_none
from ...constants import C_0, MICROMETER, RADIAN, inf
from ...exceptions import Tidy3dError, ValidationError
from ...log import log
from ..mode.mode_solver import ModeSolver

AnnotatedMedium = Annotated[MediumType, pydantic.Field(discriminator=TYPE_TAG_STR)]


EVANESCENT_TAIL = 1.5


class RectangularDielectric(Tidy3dBaseModel):
    """General rectangular dielectric waveguide

    Supports:
    - Strip and rib geometries
    - Angled sidewalls
    - Modes in waveguide bends
    - Surface and sidewall loss models
    - Coupled waveguides
    """

    wavelength: Union[float, ArrayFloat1D] = pydantic.Field(
        ...,
        title="Wavelength",
        description="Wavelength(s) at which to calculate modes (in μm).",
        units=MICROMETER,
    )

    core_width: Union[Size1D, ArrayFloat1D] = pydantic.Field(
        ...,
        title="Core width",
        description="Core width at the top of the waveguide. If set to an array, defines "
        "the widths of adjacent waveguides.",
        units=MICROMETER,
    )

    core_thickness: Size1D = pydantic.Field(
        ...,
        title="Core Thickness",
        description="Thickness of the core layer.",
        units=MICROMETER,
    )

    core_medium: MediumType = pydantic.Field(
        ...,
        title="Core Medium",
        description="Medium associated with the core layer.",
        discriminator=TYPE_TAG_STR,
    )

    clad_medium: Union[AnnotatedMedium, Tuple[AnnotatedMedium, ...]] = pydantic.Field(
        ...,
        title="Clad Medium",
        description="Medium associated with the upper cladding layer. A sequence of mediums can "
        "be used to create a layered clad.",
    )

    box_medium: Union[AnnotatedMedium, Tuple[AnnotatedMedium, ...]] = pydantic.Field(
        None,
        title="Box Medium",
        description="Medium associated with the lower cladding layer. A sequence of mediums can "
        "be used to create a layered substrate. If not set, the first clad medium is used.",
    )

    slab_thickness: Size1D = pydantic.Field(
        0.0,
        title="Slab Thickness",
        description="Thickness of the slab for rib geometry.",
        units=MICROMETER,
    )

    clad_thickness: Union[Size1D, ArrayFloat1D] = pydantic.Field(
        None,
        title="Clad Thickness",
        description="Domain size above the core layer. An array can be used to define a layered "
        "clad. The last layer is extended into the PML as an infinitely thick layer.",
        units=MICROMETER,
    )

    box_thickness: Union[Size1D, ArrayFloat1D] = pydantic.Field(
        None,
        title="Box Thickness",
        description="Domain size below the core layer. An array can be used to define a layered "
        "substrate. The last layer is extended into the PML as an infinitely thick layer.",
        units=MICROMETER,
    )

    side_margin: Size1D = pydantic.Field(
        None,
        title="Side Margin",
        description="Domain size to the sides of the waveguide core.",
        units=MICROMETER,
    )

    sidewall_angle: float = pydantic.Field(
        0.0,
        title="Sidewall Angle",
        description="Angle of the core sidewalls measured from the vertical direction (in "
        "radians).  Positive (negative) values create waveguides with bases wider (narrower) "
        "than their tops.",
        units=RADIAN,
    )

    gap: Union[float, ArrayFloat1D] = pydantic.Field(
        0.0,
        title="Gap",
        description="Distance between adjacent waveguides, measured at the top core edges.  "
        "An array can be used to define one gap per pair of adjacent waveguides.",
        units=MICROMETER,
    )

    sidewall_thickness: Size1D = pydantic.Field(
        0.0,
        title="Sidewall Thickness",
        description="Sidewall layer thickness (within core).",
        units=MICROMETER,
    )

    sidewall_medium: MediumType = pydantic.Field(
        None,
        title="Sidewall medium",
        description="Medium associated with the sidewall layer to model sidewall losses.",
        discriminator=TYPE_TAG_STR,
    )

    surface_thickness: Size1D = pydantic.Field(
        0.0,
        title="Surface Thickness",
        description="Thickness of the surface layers defined on the top of the waveguide and  "
        "slab regions (if any).",
        units=MICROMETER,
    )

    surface_medium: MediumType = pydantic.Field(
        None,
        title="Surface Medium",
        description="Medium associated with the surface layer to model surface losses.",
        discriminator=TYPE_TAG_STR,
    )

    origin: Coordinate = pydantic.Field(
        (0, 0, 0),
        title="Origin",
        description="Center of the waveguide geometry.  This coordinate represents the base "
        "of the waveguides (substrate surface) in the normal axis, and center of the geometry "
        "in the remaining axes.",
        units=MICROMETER,
    )

    length: Size1D = pydantic.Field(
        1e30,
        title="Length",
        description="Length of the waveguides in the propagation direction",
        units=MICROMETER,
    )

    propagation_axis: Axis = pydantic.Field(
        0,
        title="Propagation Axis",
        description="Axis of propagation of the waveguide",
    )

    normal_axis: Axis = pydantic.Field(
        2,
        title="Normal Axis",
        description="Axis normal to the substrate surface",
    )

    mode_spec: ModeSpec = pydantic.Field(
        ModeSpec(num_modes=2),
        title="Mode Specification",
        description=":class:`ModeSpec` defining waveguide mode properties.",
    )

    grid_resolution: int = pydantic.Field(
        15,
        title="Grid Resolution",
        description="Solver grid resolution per wavelength.",
    )

    max_grid_scaling: float = pydantic.Field(
        1.2,
        title="Maximal Grid Scaling",
        description="Maximal size increase between adjacent grid boundaries.",
    )

    @pydantic.validator("wavelength", "core_width", "gap", always=True)
    def _set_non_negative_array(cls, val):
        """Ensure values are not negative and convert to numpy arrays."""
        val = numpy.array(val, ndmin=1)
        if any(val < 0):
            raise ValidationError("Values may not be negative.")
        return val

    @pydantic.validator("gap", always=True)
    @skip_if_fields_missing(["core_width"])
    def _validate_gaps(cls, val, values):
        """Ensure the number of gaps is compatible with the number of cores supplied."""
        if val.size == 1 and values["core_width"].size != 2:
            # If a single value is defined, use it for all gaps
            return numpy.array([val[0]] * (values["core_width"].size - 1))
        if val.size != values["core_width"].size - 1:
            raise ValidationError("Number of gaps must be 1 less than number of core widths.")
        return val

    @pydantic.root_validator
    def _set_box_medium(cls, values):
        """Set BOX medium same as cladding as default value."""
        box_medium = values.get("box_medium")
        if box_medium is None:
            clad_medium = values.get("clad_medium")
            if clad_medium is None:
                return values
            if isinstance(clad_medium, tuple):
                clad_medium = clad_medium[0]
            values["box_medium"] = clad_medium
        return values

    @pydantic.root_validator
    def _set_clad_thickness(cls, values):
        """Set default clad/BOX thickness based on the max wavelength in the medium."""
        for side in ("clad", "box"):
            val = values.get(side + "_thickness")
            if val is None:
                wavelength = values.get("wavelength")
                medium = values.get(side + "_medium")
                if wavelength is None or medium is None:
                    return values
                if isinstance(medium, tuple):
                    medium = medium[0]
                n = numpy.array([medium.nk_model(f)[0] for f in C_0 / wavelength])
                lda = wavelength / n
                values[side + "_thickness"] = EVANESCENT_TAIL * lda.max()
            elif isinstance(val, float):
                if val < 0:
                    raise ValidationError("Thickness may not be negative.")
            else:
                values[side + "_thickness"] = cls._set_non_negative_array(val)
        return values

    @pydantic.root_validator
    def _validate_layers(cls, values):
        """Ensure the number of clad media is compatible with the number of layers supplied."""
        for side in ("clad", "box"):
            thickness = values.get(side + "_thickness")
            medium = values.get(side + "_medium")
            if thickness is None or medium is None:
                return values
            num_layers = 1 if isinstance(thickness, float) else thickness.size
            num_media = 1 if not isinstance(medium, tuple) else len(medium)
            if num_layers != num_media:
                raise ValidationError(
                    f"Number of '{side}_thickness' values ({num_layers}) must be equal to that of "
                    f"'{side}_medium' ({num_media})."
                )
        return values

    @pydantic.root_validator
    def _set_side_margin(cls, values):
        """Set default side margin based on BOX and cladding thicknesses."""
        clad_thickness = values.get("clad_thickness")
        box_thickness = values.get("box_thickness")
        if clad_thickness is None or box_thickness is None:
            return values
        if values["side_margin"] is None:
            if not isinstance(clad_thickness, float):
                clad_thickness = clad_thickness.sum()
            if not isinstance(box_thickness, float):
                box_thickness = box_thickness.sum()
            values["side_margin"] = max(clad_thickness, box_thickness)
        return values

    @pydantic.root_validator
    def _ensure_consistency(cls, values):
        """Ensure consistency in setting surface/sidewall models and propagation/normal axes."""
        sidewall_thickness = values["sidewall_thickness"]
        sidewall_medium = values["sidewall_medium"]
        surface_thickness = values["surface_thickness"]
        surface_medium = values["surface_medium"]
        propagation_axis = values["propagation_axis"]
        normal_axis = values["normal_axis"]

        if sidewall_thickness > 0 and sidewall_medium is None:
            raise ValidationError(
                "Sidewall medium must be provided when sidewall thickness is greater than 0."
            )

        if sidewall_thickness == 0 and sidewall_medium is not None:
            log.warning("Sidewall medium not used because sidewall thickness is zero.")

        if surface_thickness > 0 and surface_medium is None:
            raise ValidationError(
                "Surface medium must be provided when surface thickness is greater than 0."
            )

        if surface_thickness == 0 and surface_medium is not None:
            log.warning("Surface medium not used because surface thickness is zero.")

        if propagation_axis == normal_axis:
            raise ValidationError("Propagation and normal axes must be different.")

        return values

    @property
    def _clad_medium(self) -> Tuple[MediumType, ...]:
        """Normalize data type to tuple."""
        if not isinstance(self.clad_medium, tuple):
            return (self.clad_medium,)
        return self.clad_medium

    @property
    def _box_medium(self) -> Tuple[MediumType, ...]:
        """Normalize data type to tuple."""
        if not isinstance(self.box_medium, tuple):
            return (self.box_medium,)
        return self.box_medium

    @property
    def _clad_thickness(self) -> ArrayFloat1D:
        """Normalize data type to array."""
        if isinstance(self.clad_thickness, float):
            return numpy.array([self.clad_thickness])
        return self.clad_thickness

    @property
    def _box_thickness(self) -> ArrayFloat1D:
        """Normalize data type to array."""
        if isinstance(self.box_thickness, float):
            return numpy.array([self.box_thickness])
        return self.box_thickness

    @cached_property
    def lateral_axis(self) -> Axis:
        """Lateral direction axis."""
        return 3 - self.propagation_axis - self.normal_axis

    def _swap_axis(
        self, lateral_coord: Any, normal_coord: Any, propagation_coord: Any
    ) -> List[Any]:
        """Swap the model coordinates to desired axes."""
        result = [None, None, None]
        result[self.lateral_axis] = lateral_coord
        result[self.propagation_axis] = propagation_coord
        result[self.normal_axis] = normal_coord
        return result

    def _translate(
        self, lateral_coord: float, normal_coord: float, propagation_coord: float
    ) -> List[float]:
        """Swap the model coordinates to desired axes and translate to origin."""
        coordinates = self._swap_axis(lateral_coord, normal_coord, propagation_coord)
        result = [a + b for a, b in zip(self.origin, coordinates)]
        return result

    def _transform_in_plane(self, lateral_coord: float, propagation_coord: float) -> List[float]:
        """Swap the model coordinates to desired axes in the substrate plane."""
        result = self._translate(lateral_coord, 0, propagation_coord)
        _, result = Box.pop_axis(result, self.normal_axis)
        return result

    @cached_property
    def height(self) -> Size1D:
        """Domain height (size in the normal direction)."""
        return self._box_thickness.sum() + self.core_thickness + self._clad_thickness.sum()

    @cached_property
    def width(self) -> Size1D:
        """Domain width (size in the lateral direction)."""
        w = self.core_width.sum() + self.gap.sum() + 2 * self.side_margin
        if self.sidewall_angle > 0:
            w += 2 * self.core_thickness * numpy.tan(self.sidewall_angle)
        return w

    @property
    def _core_starts(self) -> List[float]:
        """Starting positions of each waveguide (x is the position in the lateral direction)."""
        core_x = [-0.5 * (self.core_width.sum() + self.gap.sum())]
        core_x.extend(core_x[0] + numpy.cumsum(self.core_width[:-1]) + numpy.cumsum(self.gap))
        return core_x

    @property
    def _override_structures(self) -> List[Structure]:
        """Build override structures to define the simulation grid."""

        # Grid resolution factor applied to the materials (increase for waveguide corners
        # and decrease for evanescent tail regions).
        scale_factor = 1.5

        freqs = C_0 / self.wavelength
        nk_core = numpy.array([self.core_medium.nk_model(f) for f in freqs])
        # Dimensions: (medium, wavelength, n/k)
        nk_clad = numpy.array([[m.nk_model(f) for f in freqs] for m in self._clad_medium])
        nk_box = numpy.array([[m.nk_model(f) for f in freqs] for m in self._box_medium])
        lda_core = self.wavelength / nk_core[:, 0]
        lda_clad = self.wavelength / nk_clad[:, :, 0]
        lda_box = self.wavelength / nk_box[:, :, 0]

        core_x = self._core_starts

        i = numpy.argmin(lda_core)
        hi_index = Medium.from_nk(n=nk_core[i, 0] * scale_factor, k=0, freq=freqs[i])

        m, i = numpy.unravel_index(numpy.argmin(lda_box), lda_box.shape)
        lo_index_box = Medium.from_nk(
            n=max(1.0, nk_box[m, i, 0] / scale_factor), k=0, freq=freqs[i]
        )

        m, i = numpy.unravel_index(numpy.argmin(lda_clad), lda_clad.shape)
        lo_index_clad = Medium.from_nk(
            n=max(1.0, nk_clad[m, i, 0] / scale_factor), k=0, freq=freqs[i]
        )

        # Gather all waveguide edge intervals into `hi_res` list
        corner_margin = max(lda_box.max(), lda_clad.max()) / self.grid_resolution
        if self.sidewall_angle > 0:
            dx = (self.core_thickness - self.slab_thickness) * numpy.tan(self.sidewall_angle)
            hi_res = [
                pair
                for x, w in zip(core_x, self.core_width)
                for pair in [
                    [x - dx - corner_margin, x + corner_margin],
                    [x + w - corner_margin, x + w + dx + corner_margin],
                ]
            ]
        elif self.sidewall_angle < 0:
            dx = (self.core_thickness - self.slab_thickness) * numpy.tan(self.sidewall_angle)
            hi_res = [
                pair
                for x, w in zip(core_x, self.core_width)
                for pair in [
                    [x - corner_margin, x - dx + corner_margin],
                    [x + w + dx - corner_margin, x + w + corner_margin],
                ]
            ]
        else:
            hi_res = [
                pair
                for x, w in zip(core_x, self.core_width)
                for pair in [
                    [x - corner_margin, x + corner_margin],
                    [x + w - corner_margin, x + w + corner_margin],
                ]
            ]

        # The gaps between waveguides can be small enough to merge adjacent high
        # resolution intervals (specially with angled sidewalls), so we merge
        # intervals that overlap
        i = 0
        while i < len(hi_res) - 1:
            if hi_res[i][1] >= hi_res[i + 1][0]:
                hi_res[i][1] = hi_res.pop(i + 1)[1]
            else:
                i += 1

        # Create override structures to improve the mode solver grid.  We want
        # high resolution around all core edges, but not along the whole slab
        # in case of rib geometry.  Start with the 4 corners:
        override_structures = [
            Structure(
                geometry=Box(
                    center=self._translate(0.5 * (a + b), y, 0),
                    size=self._swap_axis(b - a, 2 * corner_margin, inf),
                ),
                medium=hi_index,
            )
            for (a, b) in hi_res
            for y in (self.slab_thickness, self.core_thickness)
        ]
        # Low resolution on the sides:
        override_structures.extend(
            Structure(
                geometry=Box(
                    center=self._translate(0.5 * (a + b), 0, 0),
                    size=self._swap_axis(b - a, inf, inf),
                ),
                medium=lo_index_clad,
            )
            for (a, b) in ((-self.width, hi_res[0][0]), (hi_res[-1][1], self.width))
        )
        # Low resolution above and below:
        override_structures.extend(
            Structure(
                geometry=Box(
                    center=self._translate(0, 0.5 * (a + b), 0),
                    size=self._swap_axis(inf, b - a, inf),
                ),
                medium=lo_index,
            )
            for (a, b, lo_index) in (
                (-2 * self._box_thickness.sum(), -corner_margin, lo_index_box),
                (
                    self.core_thickness + corner_margin,
                    self.core_thickness + 2 * self._clad_thickness.sum(),
                    lo_index_clad,
                ),
            )
        )

        return override_structures

    @cached_property
    def grid_spec(self) -> GridSpec:
        """Waveguide grid specification with overriding geometry."""
        grid_spec = GridSpec.auto(
            min_steps_per_wvl=self.grid_resolution,
            wavelength=self.wavelength.min(),
            override_structures=self._override_structures,
            max_scale=self.max_grid_scaling,
        )
        return grid_spec

    @cached_property
    def structures(self) -> List[Structure]:
        """Waveguide structures for simulation, including the core(s), slabs (if any), and bottom
        cladding, if different from the top. For bend modes, the structure is a 270 degree bend
        regardless of :attr:`length`."""

        structures = []

        # Lower cladding
        thicknesses = self._box_thickness.tolist()
        # Make sure the last layer extends into the PML
        thicknesses[-1] += 2.0 * self.wavelength.max()
        total = 0
        for medium, thickness in zip(self._box_medium, thicknesses):
            structures.append(
                Structure(
                    geometry=Box(
                        center=self._translate(0, -total - thickness / 2, 0),
                        size=self._swap_axis(inf, thickness, self.length),
                    ),
                    medium=medium,
                )
            )
            total += thickness

        # Upper cladding
        thicknesses = self._clad_thickness.tolist()
        thicknesses[0] += self.core_thickness
        # Make sure the last layer extends into the PML
        thicknesses[-1] += 2.0 * self.wavelength.max()
        total = 0
        for medium, thickness in zip(self._clad_medium, thicknesses):
            structures.append(
                Structure(
                    geometry=Box(
                        center=self._translate(0, total + thickness / 2, 0),
                        size=self._swap_axis(inf, thickness, self.length),
                    ),
                    medium=medium,
                )
            )
            total += thickness

        # Create a local copy of these values, as they will be modified
        # according to the desired geometry
        core_w = numpy.array(self.core_width, copy=True)
        core_t = self.core_thickness
        slab_t = self.slab_thickness

        core_x = self._core_starts

        if self.mode_spec.bend_radius is None or self.mode_spec.bend_radius == 0.0:
            half_length = 0.5 * self.length

            def polyslab_vertices(x, w):
                return (
                    self._transform_in_plane(x, -half_length),
                    self._transform_in_plane(x + w, -half_length),
                    self._transform_in_plane(x + w, half_length),
                    self._transform_in_plane(x, half_length),
                )

        else:
            if (self.normal_axis > self.lateral_axis) != (self.mode_spec.bend_axis == 1):
                raise Tidy3dError(
                    "Waveguide band axis must be the substrate normal "
                    f"(mode_spec.bend_axis = {1 - self.mode_spec.bend_axis})"
                )

            bend_radius = self.mode_spec.bend_radius

            # 10 nm resolution (at center)
            num_points = 1 + int(0.5 + 1.5 * numpy.pi * abs(bend_radius) / 0.01)

            angles = numpy.linspace(-0.75 * numpy.pi, 0.75 * numpy.pi, num_points)
            if bend_radius < 0:
                angles = -angles
            sin = numpy.sin(angles)
            cos = numpy.cos(angles)

            def polyslab_vertices(x, w):
                r_in = bend_radius + x
                v_in = numpy.vstack((-bend_radius + r_in * cos, r_in * sin)).T
                r_out = r_in + w
                v_out = numpy.vstack((-bend_radius + r_out * cos, r_out * sin)).T
                return [self._transform_in_plane(*v) for v in list(v_out) + list(v_in[::-1])]

        # Create the actual waveguide geometry

        normal_origin = self.origin[self.normal_axis]

        # Surface and sidewall loss regions are created first, so that the core
        # can be applied on top.
        if self.surface_thickness > 0:
            structures.extend(
                Structure(
                    geometry=PolySlab(
                        vertices=polyslab_vertices(x, w),
                        slab_bounds=(
                            normal_origin + core_t - self.surface_thickness,
                            normal_origin + core_t,
                        ),
                        sidewall_angle=self.sidewall_angle,
                        reference_plane="top",
                        axis=self.normal_axis,
                    ),
                    medium=self.surface_medium,
                )
                for x, w in zip(core_x, core_w)
            )

            # Add loss region over slab surface, if rib geometry
            if slab_t > 0:
                structures.append(
                    Structure(
                        geometry=Box(
                            center=self._translate(0, 0.5 * slab_t, 0),
                            size=self._swap_axis(inf, slab_t, self.length),
                        ),
                        medium=self.surface_medium,
                    )
                )

            # Correct core geometry to leave the lossy regions with their
            # specified thickness
            dx = self.surface_thickness * numpy.tan(self.sidewall_angle)
            core_x -= dx
            core_w += 2 * dx
            core_t -= self.surface_thickness
            slab_t = max(0, slab_t - self.surface_thickness)

        if self.sidewall_thickness > 0:
            structures.extend(
                Structure(
                    geometry=PolySlab(
                        vertices=polyslab_vertices(x, w),
                        slab_bounds=(normal_origin, normal_origin + core_t),
                        sidewall_angle=self.sidewall_angle,
                        reference_plane="top",
                        axis=self.normal_axis,
                    ),
                    medium=self.sidewall_medium,
                )
                for x, w in zip(core_x, core_w)
            )

            # Core position offset and width reduction to accommodate lossy
            # regions
            dx = self.sidewall_thickness / numpy.cos(self.sidewall_angle)
            core_x += dx
            core_w -= 2 * dx

        # Waveguide cores
        structures.extend(
            Structure(
                geometry=PolySlab(
                    vertices=polyslab_vertices(x, w),
                    slab_bounds=(normal_origin, normal_origin + core_t),
                    sidewall_angle=self.sidewall_angle,
                    reference_plane="top",
                    axis=self.normal_axis,
                ),
                medium=self.core_medium,
            )
            for x, w in zip(core_x, core_w)
        )

        # Slab for rib geometry
        if slab_t > 0:
            structures.append(
                Structure(
                    geometry=Box(
                        center=self._translate(0, 0.5 * slab_t, 0),
                        size=self._swap_axis(inf, slab_t, self.length),
                    ),
                    medium=self.core_medium,
                )
            )

        return structures

    @cached_property
    def mode_solver(self) -> ModeSolver:
        """Create a mode solver based on this waveguide structure

        Returns
        -------
        :class:`ModeSolver`

        Example
        -------
        >>> import tidy3d.plugins.waveguide as waveguide
        >>> wg = waveguide.RectangularDielectric(
        ...     wavelength=1.55,
        ...     core_width=0.5,
        ...     core_thickness=0.22,
        ...     core_medium=Medium(permittivity=3.48**2),
        ...     clad_medium=Medium(permittivity=1.45**2),
        ...     # num_modes=2,
        ... )
        >>> mode_data = wg.mode_solver.solve()  # doctest: +SKIP
        >>> mode_data.n_eff.values  # doctest: +SKIP
        array([[2.4554183, 1.7839471]], dtype=float32)

        """
        freqs = C_0 / self.wavelength
        f_max = freqs.max()
        f_min = freqs.min()
        freq0 = 0.5 * (f_max + f_min)
        fwidth = max(0.1 * freq0, f_max - f_min)

        plane = Box(
            center=self._translate(0, 0.5 * self.height - self._box_thickness.sum(), 0),
            size=self._swap_axis(self.width, self.height, 0),
        )

        # Source used only to silence warnings
        mode_source = ModeSource(
            center=plane.center,
            size=plane.size,
            source_time=GaussianPulse(freq0=freq0, fwidth=fwidth),
            direction="+",
            mode_spec=self.mode_spec,
        )

        simulation = Simulation(
            center=plane.center,
            size=plane.size,
            medium=self._clad_medium[-1],
            structures=self.structures,
            boundary_spec=BoundarySpec.all_sides(Periodic()),
            grid_spec=self.grid_spec,
            sources=[mode_source],
            run_time=1e-12,
        )

        mode_solver = ModeSolver(
            simulation=simulation,
            plane=plane,
            mode_spec=self.mode_spec,
            freqs=freqs,
        )

        return mode_solver

    @property
    def n_eff(self) -> ModeIndexDataArray:
        """Calculate the effective index."""
        return self.mode_solver.data.n_eff

    @property
    def n_complex(self) -> ModeIndexDataArray:
        """Calculate the complex effective index."""
        return self.mode_solver.data.n_complex

    @property
    def n_group(self) -> ModeIndexDataArray:
        r"""Calculate the group index."""
        return self.mode_solver.data.n_group

    @property
    def mode_area(self) -> FreqModeDataArray:
        """Calculate the effective mode area."""
        return self.mode_solver.data.mode_area

    # plot wrappers

    def plot(
        self,
        x: float = None,
        y: float = None,
        z: float = None,
        ax: Ax = None,
        source_alpha: float = None,
        monitor_alpha: float = None,
        **patch_kwargs,
    ) -> Ax:
        """Plot each of simulation's components on a plane defined by one nonzero x,y,z coordinate.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        source_alpha : float = None
            Opacity of the sources. If ``None``, uses Tidy3d default.
        monitor_alpha : float = None
            Opacity of the monitors. If ``None``, uses Tidy3d default.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """
        return self.mode_solver.simulation.plot(
            x=x,
            y=y,
            z=z,
            ax=ax,
            source_alpha=source_alpha,
            monitor_alpha=monitor_alpha,
            **patch_kwargs,
        )

    def plot_eps(
        self,
        x: float = None,
        y: float = None,
        z: float = None,
        freq: float = None,
        alpha: float = None,
        source_alpha: float = None,
        monitor_alpha: float = None,
        ax: Ax = None,
    ) -> Ax:
        """Plot each of simulation's components on a plane defined by one nonzero x,y,z coordinate.
        The permittivity is plotted in grayscale based on its value at the specified frequency.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        freq : float = None
            Frequency to evaluate the relative permittivity of all mediums.
            If not specified, evaluates at infinite frequency.
        alpha : float = None
            Opacity of the structures being plotted.
            Defaults to the structure default alpha.
        source_alpha : float = None
            Opacity of the sources. If ``None``, uses Tidy3d default.
        monitor_alpha : float = None
            Opacity of the monitors. If ``None``, uses Tidy3d default.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """
        return self.mode_solver.simulation.plot_eps(
            x=x,
            y=y,
            z=z,
            freq=freq,
            alpha=alpha,
            source_alpha=source_alpha,
            monitor_alpha=monitor_alpha,
            ax=ax,
        )

    def plot_structures(
        self, x: float = None, y: float = None, z: float = None, ax: Ax = None
    ) -> Ax:
        """Plot each of simulation's structures on a plane defined by one nonzero x,y,z coordinate.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """
        return self.mode_solver.simulation.plot_structures(
            x=x,
            y=y,
            z=z,
            ax=ax,
        )

    def plot_structures_eps(
        self,
        x: float = None,
        y: float = None,
        z: float = None,
        freq: float = None,
        alpha: float = None,
        cbar: bool = True,
        reverse: bool = False,
        ax: Ax = None,
    ) -> Ax:
        """Plot each of simulation's structures on a plane defined by one nonzero x,y,z coordinate.
        The permittivity is plotted in grayscale based on its value at the specified frequency.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        freq : float = None
            Frequency to evaluate the relative permittivity of all mediums.
            If not specified, evaluates at infinite frequency.
        reverse : bool = False
            If ``False``, the highest permittivity is plotted in black.
            If ``True``, it is plotteed in white (suitable for black backgrounds).
        cbar : bool = True
            Whether to plot a colorbar for the relative permittivity.
        alpha : float = None
            Opacity of the structures being plotted.
            Defaults to the structure default alpha.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """
        return self.mode_solver.simulation.plot_structures_eps(
            x=x,
            y=y,
            z=z,
            freq=freq,
            alpha=alpha,
            cbar=cbar,
            reverse=reverse,
            ax=ax,
        )

    def plot_grid(
        self,
        x: float = None,
        y: float = None,
        z: float = None,
        ax: Ax = None,
        **kwargs,
    ) -> Ax:
        """Plot the cell boundaries as lines on a plane defined by one nonzero x,y,z coordinate.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.
        **kwargs
            Optional keyword arguments passed to the matplotlib ``LineCollection``.
            For details on accepted values, refer to
            `Matplotlib's documentation <https://tinyurl.com/2p97z4cn>`_.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """
        return self.mode_solver.simulation.plot_grid(
            x=x,
            y=y,
            z=z,
            ax=ax,
            **kwargs,
        )

    @add_ax_if_none
    def plot_geometry_edges(
        self,
        color: str = "k",
        ax: Ax = None,
    ) -> Ax:
        """Plot the waveguide cross-section geometry edges.

        Parameters
        ----------
        color : Color to use for the geometry edges.
        ax : matplotlib.axes._subplots.Axes = None
            matplotlib axes to plot on, if not specified, one is created.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """
        kwargs = {"color": color, "linewidth": pyplot.rcParams["grid.linewidth"]}
        if self.normal_axis < self.lateral_axis:
            x0 = self.origin[self.lateral_axis]
            ax.axvline(x0, **kwargs)
            x = x0
            for t in self._box_thickness:
                x -= t
                ax.axvline(x, **kwargs)
            x = x0 + self.core_thickness
            for t in self._clad_thickness:
                x += t
                ax.axvline(x, **kwargs)
        else:
            y0 = self.origin[self.normal_axis]
            ax.axhline(y0, **kwargs)
            y = y0
            for t in self._box_thickness:
                y -= t
                ax.axhline(y, **kwargs)
            y = y0 + self.core_thickness
            for t in self._clad_thickness:
                y += t
                ax.axhline(y, **kwargs)

        dx = self.core_thickness * numpy.tan(self.sidewall_angle)
        u = min(self.lateral_axis, self.normal_axis)
        v = max(self.lateral_axis, self.normal_axis)
        for x0, w in zip(self._core_starts, self.core_width):
            plot_x = []
            plot_y = []
            for x, y in zip(
                [x0 - dx, x0, x0 + w, x0 + w + dx],
                [0, self.core_thickness, self.core_thickness, 0],
            ):
                p = self._translate(x, y, 0)
                plot_x.append(p[u])
                plot_y.append(p[v])
            ax.plot(plot_x, plot_y, linestyle="-", **kwargs)
        ax.set_aspect("equal")
        return ax

    def plot_field(
        self,
        field_name: str,
        val: Literal["real", "imag", "abs"] = "real",
        eps_alpha: float = 0.2,
        robust: bool = True,
        vmin: float = None,
        vmax: float = None,
        ax: Ax = None,
        geometry_edges: str = None,
        **sel_kwargs,
    ) -> Ax:
        """Plot the field for a :class:`.ModeSolverData` with :class:`.Simulation` plot overlaid.

        Parameters
        ----------
        field_name : str
            Name of ``field`` component to plot (eg. ``'Ex'``).
            Also accepts ``'E'`` and ``'H'`` to plot the vector magnitudes of the electric and
            magnetic fields, and ``'S'`` for the Poynting vector.
        val : Literal['real', 'imag', 'abs', 'abs^2', 'dB'] = 'real'
            Which part of the field to plot.
        eps_alpha : float = 0.2
            Opacity of the structure permittivity.
            Must be between 0 and 1 (inclusive).
        robust : bool = True
            If True and vmin or vmax are absent, uses the 2nd and 98th percentiles of the data
            to compute the color limits. This helps in visualizing the field patterns especially
            in the presence of a source.
        vmin : float = None
            The lower bound of data range that the colormap covers. If ``None``, they are
            inferred from the data and other keyword arguments.
        vmax : float = None
            The upper bound of data range that the colormap covers. If ``None``, they are
            inferred from the data and other keyword arguments.
        ax : matplotlib.axes._subplots.Axes = None
            matplotlib axes to plot on, if not specified, one is created.
        geometry_edges : Optional color to use for the geometry edges overlaid on the fields.
        sel_kwargs : keyword arguments used to perform ``.sel()`` selection in the monitor data.
            These kwargs can select over the spatial dimensions (``x``, ``y``, ``z``),
            frequency or time dimensions (``f``, ``t``) or `mode_index`, if applicable.
            For the plotting to work appropriately, the resulting data after selection must contain
            only two coordinates with len > 1.
            Furthermore, these should be spatial coordinates (``x``, ``y``, or ``z``).

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """
        ax = self.mode_solver.plot_field(
            field_name=field_name,
            val=val,
            eps_alpha=eps_alpha,
            robust=robust,
            vmin=vmin,
            vmax=vmax,
            ax=ax,
            **sel_kwargs,
        )
        if geometry_edges is not None:
            self.plot_geometry_edges(geometry_edges, ax=ax)
        return ax



"""EME dataset"""

from __future__ import annotations

import pydantic.v1 as pd

from ...data.data_array import (
    EMECoefficientDataArray,
    EMEModeIndexDataArray,
    EMEScalarFieldDataArray,
    EMEScalarModeFieldDataArray,
    EMESMatrixDataArray,
)
from ...data.dataset import Dataset, ElectromagneticFieldDataset


class EMESMatrixDataset(Dataset):
    """Dataset storing S matrix."""

    S11: EMESMatrixDataArray = pd.Field(
        ...,
        title="S11 matrix",
        description="S matrix relating output modes at port 1 to input modes at port 1.",
    )
    S12: EMESMatrixDataArray = pd.Field(
        ...,
        title="S12 matrix",
        description="S matrix relating output modes at port 1 to input modes at port 2.",
    )
    S21: EMESMatrixDataArray = pd.Field(
        ...,
        title="S21 matrix",
        description="S matrix relating output modes at port 2 to input modes at port 1.",
    )
    S22: EMESMatrixDataArray = pd.Field(
        ...,
        title="S22 matrix",
        description="S matrix relating output modes at port 2 to input modes at port 2.",
    )


class EMECoefficientDataset(Dataset):
    """Dataset storing expansion coefficients for the modes in a cell.
    These are defined at the cell centers.
    """

    A: EMECoefficientDataArray = pd.Field(
        ...,
        title="A coefficient",
        description="Coefficient for forward mode in this cell.",
    )
    B: EMECoefficientDataArray = pd.Field(
        ...,
        title="B coefficient",
        description="Coefficient for backward mode in this cell.",
    )


class EMEFieldDataset(ElectromagneticFieldDataset):
    """Dataset storing scalar components of E and H fields as a function of freq, mode_index, and port_index."""

    Ex: EMEScalarFieldDataArray = pd.Field(
        None,
        title="Ex",
        description="Spatial distribution of the x-component of the electric field of the mode.",
    )
    Ey: EMEScalarFieldDataArray = pd.Field(
        None,
        title="Ey",
        description="Spatial distribution of the y-component of the electric field of the mode.",
    )
    Ez: EMEScalarFieldDataArray = pd.Field(
        None,
        title="Ez",
        description="Spatial distribution of the z-component of the electric field of the mode.",
    )
    Hx: EMEScalarFieldDataArray = pd.Field(
        None,
        title="Hx",
        description="Spatial distribution of the x-component of the magnetic field of the mode.",
    )
    Hy: EMEScalarFieldDataArray = pd.Field(
        None,
        title="Hy",
        description="Spatial distribution of the y-component of the magnetic field of the mode.",
    )
    Hz: EMEScalarFieldDataArray = pd.Field(
        None,
        title="Hz",
        description="Spatial distribution of the z-component of the magnetic field of the mode.",
    )


class EMEModeSolverDataset(ElectromagneticFieldDataset):
    """Dataset storing EME modes as a function of freq, mode_index, and cell_index."""

    n_complex: EMEModeIndexDataArray = pd.Field(
        ...,
        title="Propagation Index",
        description="Complex-valued effective propagation constants associated with the mode.",
    )

    Ex: EMEScalarModeFieldDataArray = pd.Field(
        ...,
        title="Ex",
        description="Spatial distribution of the x-component of the electric field of the mode.",
    )
    Ey: EMEScalarModeFieldDataArray = pd.Field(
        ...,
        title="Ey",
        description="Spatial distribution of the y-component of the electric field of the mode.",
    )
    Ez: EMEScalarModeFieldDataArray = pd.Field(
        ...,
        title="Ez",
        description="Spatial distribution of the z-component of the electric field of the mode.",
    )
    Hx: EMEScalarModeFieldDataArray = pd.Field(
        ...,
        title="Hx",
        description="Spatial distribution of the x-component of the magnetic field of the mode.",
    )
    Hy: EMEScalarModeFieldDataArray = pd.Field(
        ...,
        title="Hy",
        description="Spatial distribution of the y-component of the magnetic field of the mode.",
    )
    Hz: EMEScalarModeFieldDataArray = pd.Field(
        ...,
        title="Hz",
        description="Spatial distribution of the z-component of the magnetic field of the mode.",
    )



"""utilities for plotting"""

from __future__ import annotations

from functools import wraps
from html import escape
from typing import Any

import matplotlib.pyplot as plt
import pydantic.v1 as pd
from matplotlib.patches import ArrowStyle, PathPatch
from matplotlib.path import Path
from numpy import array, concatenate, inf, ones

from ..exceptions import SetupError
from .base import Tidy3dBaseModel
from .types import Ax

""" Constants """

# add this around extents of plots
PLOT_BUFFER = 0.3

ARROW_COLOR_MONITOR = "orange"
ARROW_COLOR_SOURCE = "green"
ARROW_COLOR_POLARIZATION = "brown"
ARROW_ALPHA = 0.8

# Arrow length in inches
ARROW_LENGTH = 0.3

""" Decorators """


def make_ax() -> Ax:
    """makes an empty ``ax``."""
    _, ax = plt.subplots(1, 1, tight_layout=True)
    return ax


def add_ax_if_none(plot):
    """Decorates ``plot(*args, **kwargs, ax=None)`` function.
    if ax=None in the function call, creates an ax and feeds it to rest of function.
    """

    @wraps(plot)
    def _plot(*args, **kwargs) -> Ax:
        """New plot function using a generated ax if None."""
        if kwargs.get("ax") is None:
            ax = make_ax()
            kwargs["ax"] = ax
        return plot(*args, **kwargs)

    return _plot


def equal_aspect(plot):
    """Decorates a plotting function returning a matplotlib axes.
    Ensures the aspect ratio of the returned axes is set to equal.
    Useful for 2D plots, like sim.plot() or sim_data.plot_fields()
    """

    @wraps(plot)
    def _plot(*args, **kwargs) -> Ax:
        """New plot function with equal aspect ratio axes returned."""
        ax = plot(*args, **kwargs)
        ax.set_aspect("equal")
        return ax

    return _plot


""" plot parameters """


class PlotParams(Tidy3dBaseModel):
    """Stores plotting parameters / specifications for a given model."""

    alpha: Any = pd.Field(1.0, title="Opacity")
    edgecolor: Any = pd.Field(None, title="Edge Color", alias="ec")
    facecolor: Any = pd.Field(None, title="Face Color", alias="fc")
    fill: bool = pd.Field(True, title="Is Filled")
    hatch: str = pd.Field(None, title="Hatch Style")
    zorder: float = pd.Field(None, title="Display Order")
    linewidth: pd.NonNegativeFloat = pd.Field(1, title="Line Width", alias="lw")

    def include_kwargs(self, **kwargs) -> PlotParams:
        """Update the plot params with supplied kwargs."""
        update_dict = {
            key: value
            for key, value in kwargs.items()
            if key not in ("type",) and value is not None and key in self.__fields__
        }
        return self.copy(update=update_dict)

    def to_kwargs(self) -> dict:
        """Export the plot parameters as kwargs dict that can be supplied to plot function."""
        kwarg_dict = self.dict()
        for ignore_key in ("type", "attrs"):
            kwarg_dict.pop(ignore_key)
        return kwarg_dict


# defaults for different tidy3d objects
plot_params_geometry = PlotParams()
plot_params_structure = PlotParams()
plot_params_source = PlotParams(alpha=0.4, facecolor="limegreen", edgecolor="limegreen", lw=3)
plot_params_monitor = PlotParams(alpha=0.4, facecolor="orange", edgecolor="orange", lw=3)
plot_params_pml = PlotParams(alpha=0.7, facecolor="gray", edgecolor="gray", hatch="x", zorder=inf)
plot_params_pec = PlotParams(alpha=1.0, facecolor="gold", edgecolor="black", zorder=inf)
plot_params_pmc = PlotParams(alpha=1.0, facecolor="lightsteelblue", edgecolor="black", zorder=inf)
plot_params_bloch = PlotParams(alpha=1.0, facecolor="orchid", edgecolor="black", zorder=inf)
plot_params_symmetry = PlotParams(edgecolor="gray", facecolor="gray", alpha=0.6, zorder=inf)
plot_params_override_structures = PlotParams(
    linewidth=0.4, edgecolor="black", fill=False, zorder=inf
)
plot_params_fluid = PlotParams(facecolor="white", edgecolor="lightsteelblue", lw=0.4, hatch="xx")
plot_params_grid = PlotParams(edgecolor="black", lw=0.2)
plot_params_lumped_element = PlotParams(
    alpha=0.4, facecolor="mediumblue", edgecolor="mediumblue", lw=3
)

# stores color of simulation.structures for given index in simulation.medium_map
MEDIUM_CMAP = [
    "#689DBC",
    "#D0698E",
    "#5E6EAD",
    "#C6224E",
    "#BDB3E2",
    "#9EC3E0",
    "#616161",
    "#877EBC",
]

# colormap for structure's permittivity in plot_eps
STRUCTURE_EPS_CMAP = "gist_yarg"
STRUCTURE_HEAT_COND_CMAP = "gist_yarg"

# default arrow style
arrow_style = ArrowStyle.Simple(head_length=12, head_width=9, tail_width=4)


"""=================================================================================================
Descartes modified from https://pypi.org/project/descartes/ for Shapely >= 1.8.0

Copyright Flexcompute 2022

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


class Polygon:
    """Adapt Shapely polygons to a common interface"""

    def __init__(self, context):
        if isinstance(context, dict):
            self.context = context["coordinates"]
        else:
            self.context = context

    @property
    def exterior(self):
        """Get polygon exterior."""
        return getattr(self.context, "exterior", None) or self.context[0]

    @property
    def interiors(self):
        """Get polygon interiors."""
        value = getattr(self.context, "interiors", None)
        if value is None:
            value = self.context[1:]
        return value


def polygon_path(polygon):
    """Constructs a compound matplotlib path from a Shapely or GeoJSON-like
    geometric object"""

    def coding(obj):
        # The codes will be all "LINETO" commands, except for "MOVETO"s at the
        # beginning of each subpath
        n = len(getattr(obj, "coords", None) or obj)
        vals = ones(n, dtype=Path.code_type) * Path.LINETO
        vals[0] = Path.MOVETO
        return vals

    ptype = polygon.geom_type
    if ptype == "Polygon":
        polygon = [Polygon(polygon)]
    elif ptype == "MultiPolygon":
        polygon = [Polygon(p) for p in polygon.geoms]

    vertices = concatenate(
        [
            concatenate(
                [array(t.exterior.coords)[:, :2]] + [array(r.coords)[:, :2] for r in t.interiors]
            )
            for t in polygon
        ]
    )
    codes = concatenate(
        [concatenate([coding(t.exterior)] + [coding(r) for r in t.interiors]) for t in polygon]
    )

    return Path(vertices, codes)


def polygon_patch(polygon, **kwargs):
    """Constructs a matplotlib patch from a geometric object

    The ``polygon`` may be a Shapely or GeoJSON-like object with or without holes.
    The ``kwargs`` are those supported by the matplotlib.patches.Polygon class
    constructor. Returns an instance of matplotlib.patches.PathPatch.

    Example
    -------
    >>> b = Point(0, 0).buffer(1.0) # doctest: +SKIP
    >>> patch = PolygonPatch(b, fc='blue', ec='blue', alpha=0.5) # doctest: +SKIP
    >>> axis.add_patch(patch) # doctest: +SKIP

    """
    return PathPatch(polygon_path(polygon), **kwargs)


"""End descartes modification
================================================================================================="""


def plot_sim_3d(sim, width=800, height=800) -> None:
    """Make 3D display of simulation in ipyython notebook."""

    try:
        from IPython.display import HTML, display
    except ImportError as e:
        raise SetupError(
            "3D plotting requires ipython to be installed "
            "and the code to be running on a jupyter notebook."
        ) from e

    js_code = """
        /**
        * Simulation Viewer Injector
        *
        * Monitors the document for elements being added in the form:
        *
        *    <div class="simulation-viewer" data-width="800" data-height="800" data-simulation="{...}" />
        *
        * This script will then inject an iframe to the viewer application, and pass it the simulation data
        * via the postMessage API on request. The script may be safely included multiple times, with only the
        * configuration of the first started script (e.g. viewer URL) applying.
        *
        */
        (function() {
            const TARGET_CLASS = "simulation-viewer";
            const ACTIVE_CLASS = "simulation-viewer-active";
            const VIEWER_URL = "https://feature-simulation-viewer.d3a9gfg7glllfq.amplifyapp.com/simulation-viewer";

            class SimulationViewerInjector {
                constructor() {
                    for (var node of document.getElementsByClassName(TARGET_CLASS)) {
                        this.injectViewer(node);
                    }

                    // Monitor for newly added nodes to the DOM
                    this.observer = new MutationObserver(this.onMutations.bind(this));
                    this.observer.observe(document.body, {childList: true, subtree: true});
                }

                onMutations(mutations) {
                    for (var mutation of mutations) {
                        if (mutation.type === 'childList') {
                            /**
                            * Have found that adding the element does not reliably trigger the mutation observer.
                            * It may be the case that setting content with innerHTML does not trigger.
                            *
                            * It seems to be sufficient to re-scan the document for un-activated viewers
                            * whenever an event occurs, as Jupyter triggers multiple events on cell evaluation.
                            */
                            var viewers = document.getElementsByClassName(TARGET_CLASS);
                            for (var node of viewers) {
                                this.injectViewer(node);
                            }
                        }
                    }
                }

                injectViewer(node) {
                    // (re-)check that this is a valid simulation container and has not already been injected
                    if (node.classList.contains(TARGET_CLASS) && !node.classList.contains(ACTIVE_CLASS)) {
                        // Mark node as injected, to prevent re-runs
                        node.classList.add(ACTIVE_CLASS);

                        var uuid;
                        if (window.crypto && window.crypto.randomUUID) {
                            uuid = window.crypto.randomUUID();
                        } else {
                            uuid = "" + Math.random();
                        }

                        var frame = document.createElement("iframe");
                        frame.width = node.dataset.width || 800;
                        frame.height = node.dataset.height || 800;
                        frame.src = VIEWER_URL + "?uuid=" + uuid;

                        var postMessageToViewer;
                        postMessageToViewer = event => {{
                            if(event.data.type === 'viewer' && event.data.uuid===uuid){{
                                var simulation = JSON.parse(node.dataset.simulation);
                                frame.contentWindow.postMessage({ type: 'jupyter', uuid, value: simulation}, '*');

                                // Run once only
                                window.removeEventListener('message', postMessageToViewer);
                            }}
                        }};
                        window.addEventListener(
                            'message',
                            postMessageToViewer,
                            false
                        );

                        node.appendChild(frame);
                    }
                }
            }

            if (!window.simulationViewerInjector) {
                window.simulationViewerInjector = new SimulationViewerInjector();
            }
        })();
    """
    html_code = f"""
    <div class="simulation-viewer" data-width="{escape(str(width))}" data-height="{escape(str(height))}" data-simulation="{escape(sim._json_string)}" />
    <script>
        {js_code}
    </script>
    """

    return display(HTML(html_code))



"""EME monitor data"""

from __future__ import annotations

from typing import Union

import pydantic.v1 as pd

from ...base_sim.data.monitor_data import AbstractMonitorData
from ...data.monitor_data import ElectromagneticFieldData, ModeSolverData
from ..monitor import EMECoefficientMonitor, EMEFieldMonitor, EMEModeSolverMonitor
from .dataset import EMECoefficientDataset, EMEFieldDataset, EMEModeSolverDataset


class EMEModeSolverData(ElectromagneticFieldData, EMEModeSolverDataset):
    """Data associated with an EME mode solver monitor."""

    monitor: EMEModeSolverMonitor = pd.Field(
        ...,
        title="EME Mode Solver Monitor",
        description="EME mode solver monitor associated with this data.",
    )


class EMEFieldData(ElectromagneticFieldData, EMEFieldDataset):
    """Data associated with an EME field monitor."""

    monitor: EMEFieldMonitor = pd.Field(
        ..., title="EME Field Monitor", description="EME field monitor associated with this data."
    )


class EMECoefficientData(AbstractMonitorData, EMECoefficientDataset):
    """Data associated with an EME coefficient monitor."""

    monitor: EMECoefficientMonitor = pd.Field(
        ...,
        title="EME Coefficient Monitor",
        description="EME coefficient monitor associated with this data.",
    )


EMEMonitorDataType = Union[EMEModeSolverData, EMEFieldData, EMECoefficientData, ModeSolverData]



"""Defnes information about a task"""

from abc import ABC
from datetime import datetime
from enum import Enum
from typing import Optional

import pydantic.v1 as pydantic


class TaskStatus(Enum):
    """The statuses that the task can be in."""

    INIT = "initialized"
    QUEUE = "queued"
    PRE = "preprocessing"
    RUN = "running"
    POST = "postprocessing"
    SUCCESS = "success"
    ERROR = "error"


class TaskBase(pydantic.BaseModel, ABC):
    """Base config for all task objects."""

    class Config:
        """configure class"""

        arbitrary_types_allowed = True


class ChargeType(str, Enum):
    """The payment method of task."""

    FREE = "free"
    PAID = "paid"


class TaskBlockInfo(TaskBase):
    """The block info that task will be blocked by all three features of DE,
    User limit and Insufficient balance"""

    chargeType: ChargeType = None
    maxFreeCount: int = None
    maxGridPoints: int = None
    maxTimeSteps: int = None


class TaskInfo(TaskBase):
    """General information about task."""

    taskId: str
    taskName: str = None
    nodeSize: int = None
    completedAt: Optional[datetime] = None
    status: str = None
    realCost: float = None
    timeSteps: int = None
    solverVersion: str = None
    createAt: Optional[datetime] = None
    estCostMin: float = None
    estCostMax: float = None
    realFlexUnit: float = None
    oriRealFlexUnit: float = None
    estFlexUnit: float = None
    estFlexCreditTimeStepping: float = None
    estFlexCreditPostProcess: float = None
    estFlexCreditMode: float = None
    s3Storage: float = None
    startSolverTime: Optional[datetime] = None
    finishSolverTime: Optional[datetime] = None
    totalSolverTime: int = None
    callbackUrl: str = None
    taskType: str = None
    metadataStatus: str = None
    taskBlockInfo: TaskBlockInfo = None


class RunInfo(TaskBase):
    """Information about the run."""

    perc_done: pydantic.confloat(ge=0.0, le=100.0)
    field_decay: pydantic.confloat(ge=0.0, le=1.0)

    def display(self):
        """Print some info."""
        print(f" - {self.perc_done:.2f} (%) done")
        print(f" - {self.field_decay:.2e} field decay from max")
